from __future__ import annotations

import json
import logging

from schematools.contrib.django.models import Dataset
from schematools.exceptions import DatasetNotFound
from schematools.loaders import CachedSchemaLoader, SchemaLoader
from schematools.naming import to_snake_case
from schematools.types import DatasetSchema, DatasetTableSchema

logger = logging.getLogger(__name__)


class DatabaseSchemaLoader(SchemaLoader):
    """A schema loader that retrieves datasets that are imported into the Django database.

    When objects are directly retrieved using ``Dataset.objects.get()``,
    there is no linking of schemas between different schema's.
    This loader interface makes sure schema's can resolve their relations by fetching
    the relevant definitions from the database.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # A collection of all instances that were retrieved using this loader.
        # This shared (and cached) instance makes sure datasets can retrieve their relations.
        self.dataset_collection = CachedSchemaLoader(self)

    def get_all_datasets(self, enable_db=None) -> dict[str, DatasetSchema]:
        """Retrieve all datasets from the database storage"""
        queryset = Dataset.objects.all()
        if enable_db is not None:
            queryset = queryset.filter(enable_db=enable_db)

        return {
            path: self._as_dataset(schema_data, view_sql)
            for path, schema_data, view_sql in queryset.values_list(
                "path", "schema_data", "view_data"
            )
        }

    def get_dataset(self, dataset_id: str, prefetch_related: bool = False) -> DatasetSchema:
        """Retrieve a single dataset from the database storage."""
        queryset = Dataset.objects.filter(name=to_snake_case(dataset_id))
        try:
            schema_data = queryset.values_list("schema_data", flat=True)[0]
            view_sql = queryset.values_list("view_data", flat=True)[0]
        except IndexError:
            raise DatasetNotFound(f"Dataset `{dataset_id}` not found.") from None

        # The dataset is connected to this collection so it can resolve relations:
        return self._as_dataset(schema_data, view_sql)

    def _as_dataset(self, schema_data: str, view_sql: str | None = None) -> DatasetSchema:
        """Convert the retrieved schema into a real object that can resolve its relations."""
        return DatasetSchema(
            json.loads(schema_data), view_sql, dataset_collection=self.dataset_collection
        )

    def get_table(self, dataset: DatasetSchema, table_ref) -> DatasetTableSchema:
        """Datasets have their tables inlined, so there is no need to support this either."""
        raise NotImplementedError("DatabaseSchemaLoader don't support versioned tables")



from __future__ import annotations

from collections import ChainMap
from dataclasses import dataclass

from schematools.maps.interfaces.mapfile import serializers, types
from schematools.types import DatasetSchema, DatasetTableSchema

MapfileStr = str


@dataclass
class MapserviceLayerContext:
    """Context for generating mapfile layers from a amsterdam datatable
    and its parent dataset."""

    dataset: DatasetSchema
    datatable: DatasetTableSchema

    @property
    def name(self):
        return self.datatable.id

    @property
    def projections(self) -> list[str]:
        return [self.dataset["crs"]]

    @property
    def srid(self) -> int:
        return int(self.dataset["crs"].split(":")[-1])  # XXX replace with our impl.

    @property
    def metadata(self) -> dict:
        return {"ows_enable_request": "*"}


class MapServiceContext:
    """Context for generating mapfiles from a amsterdam dataset"""

    def __init__(self, dataset: DatasetSchema):
        self.dataset = dataset

    @property
    def name(self):
        return self.dataset.id

    @property
    def layers(self) -> list[MapserviceLayerContext]:
        return [MapserviceLayerContext(self.dataset, dt) for dt in self.dataset.tables]


@dataclass
class Generator:
    def serialize(self, dataset) -> str:
        raise NotImplementedError

    def __call__(self, dataset):
        return self.serialize(dataset)


@dataclass
class MapfileGenerator(Generator):
    serializer: serializers.MappyfileSerializer

    def generate_feature_class(self, feature_dict, base_styles=None) -> types.FeatureClass:
        feature_class = types.FeatureClass(
            name=feature_dict["name"], expression=feature_dict.get("expression")
        )
        styles = feature_dict.get("styles", base_styles)

        for index, style in enumerate(styles):
            base_style: dict = {}
            try:
                base_style = base_styles[index]
            except IndexError:
                pass
            feature_class.add_style(ChainMap(style, base_style))
        return feature_class

    def generate_layer(self, context: MapserviceLayerContext) -> types.Layer:
        return types.Layer(
            name=context.name,
            type=types.LayerType.polygon,
            with_connection=types.Connection.for_postgres("postgres", "", "postgres", "postgres"),
            data=[
                types.Data.for_postgres(
                    "geometry",
                    f"{context.dataset.id}.{context.name}",
                    srid=context.srid,
                    UNIQUE="id",
                )
            ],
            classes=[
                types.FeatureClass(
                    styles=[{"__type__": "style", "color": [200, 50, 50], "antialias": True}]
                )
            ],
            metadata=types.Metadata(context.metadata),
        )

    def serialize(self, dataset: DatasetSchema) -> MapfileStr:
        context = MapServiceContext(dataset)
        mapfile = types.Mapfile(name=context.name, layers=[], include=["header.inc"])
        mapfile.layers.extend(map(self.generate_layer, context.layers))
        return MapfileStr(self.serializer(mapfile))


@dataclass
class LegacyMapfileGenerator(Generator):
    serializer: serializers.JinjaSerializer

    def serialize(self, dataset) -> str:
        ds_dict = dataset.__dict__

        # get layers
        ds_dict["layers"] = (x.__dict__ for x in dataset.maplayer_set.all())
        ds_dict["id_field"] = dataset.pk_field

        template_file = "default.map.template"
        if dataset.map_template:
            template_file = dataset.map_template

        return self.serializer(template_file, context={"ds": ds_dict})



"""Cli tools."""

from __future__ import annotations

import io
import json
import logging
import operator
import os
import sys
from collections import defaultdict
from collections.abc import Iterable
from functools import reduce
from importlib.metadata import version
from pathlib import Path, PosixPath
from typing import Any

import click
import jsonschema
import requests
import sqlalchemy
from deepdiff import DeepDiff
from jsonschema import draft7_format_checker
from sqlalchemy import inspect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.future import create_engine
from sqlalchemy.schema import CreateTable

from schematools import (
    COMPATIBLE_METASCHEMAS,
    DEFAULT_PROFILE_URL,
    DEFAULT_SCHEMA_URL,
    ckan,
    validation,
)
from schematools.exceptions import (
    DatasetNotFound,
    DuplicateScopeId,
    IncompatibleMetaschema,
    ParserError,
    SchemaObjectNotFound,
)
from schematools.exports.csv import export_csvs
from schematools.exports.geopackage import export_geopackages
from schematools.exports.jsonlines import export_jsonls
from schematools.factories import tables_factory
from schematools.importer.base import BaseImporter
from schematools.importer.geojson import GeoJSONImporter
from schematools.importer.ndjson import NDJSONImporter
from schematools.introspect.db import introspect_db_schema
from schematools.introspect.geojson import introspect_geojson_files
from schematools.loaders import FileSystemSchemaLoader, get_schema_loader
from schematools.maps import create_mapfile
from schematools.naming import to_snake_case, toCamelCase
from schematools.permissions.db import (
    apply_schema_and_profile_permissions,
    introspect_permissions,
    revoke_permissions,
)
from schematools.provenance.create import ProvenanceIteration
from schematools.types import DatasetSchema, Publisher, Scope, SemVer

# Configure a simple stdout logger for permissions output
logger = logging.getLogger("schematools.permissions")
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
# Add simple formatting for cli useage
formatter = logging.Formatter("%(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)
pkg_version = version("amsterdam-schema-tools")

option_db_url = click.option(
    "--db-url",
    envvar="DATABASE_URL",
    required=True,
    help="DSN of database, can also use DATABASE_URL environment.",
)
option_schema_url = click.option(
    "--schema-url",
    envvar="SCHEMA_URL",
    default=DEFAULT_SCHEMA_URL,
    show_default=True,
    required=True,
    help="Url where valid amsterdam schema files are found. "
    "SCHEMA_URL can also be provided as environment variable.",
)

argument_dataset_id = click.argument(
    "dataset_id",
)

option_profile_url = click.option(
    "--profile-url",
    envvar="PROFILE_URL",
    default=DEFAULT_PROFILE_URL,
    show_default=True,
    required=True,
    help="Url where valid amsterdam profile files are found. "
    "PROFILE_URL can also be provided as environment variable.",
)

argument_profile_location = click.argument(
    "profile_location",
    metavar="(PROFILE-FILENAME | NONE)",
)

argument_role = click.argument(
    "role",
)


def _get_engine(db_url: str, pg_schemas: list[str] | None = None) -> sqlalchemy.engine.Engine:
    """Initialize the SQLAlchemy engine, and report click errors."""
    kwargs = {"pool_pre_ping": True}
    if pg_schemas is not None:
        csearch_path = ",".join(pg_schemas + ["public"])
        kwargs["connect_args"] = {"options": f"-csearch_path={csearch_path}"}
    try:
        return create_engine(db_url, **kwargs)
    except SQLAlchemyError as e:
        raise click.BadParameter(str(e), param_hint="--db-url") from e


def main() -> None:
    """Main entry point.

    This catches relevant errors, so the user is not
    confronted with internal tracebacks.
    """
    try:
        schema()
    except (OSError, SQLAlchemyError, ParserError) as e:
        click.echo(f"{e.__class__.__name__}: {e}", err=True)
        exit(1)


@click.group()
def schema() -> None:
    """Command line utility to work with Amsterdam Schema files."""


@schema.group("import")
def import_() -> None:
    """Subcommand to import data."""


@schema.group("export")
def export() -> None:
    """Subcommand to export data."""


@schema.group("tocase")
def tocase() -> None:
    """Subcommand to make case-changes."""


@schema.group()
def show() -> None:
    """Show existing metadata."""


@schema.group()
def permissions() -> None:
    """Subcommand for permissions."""


@schema.group()
def kafka() -> None:
    """Subcommand to consume or produce kafka events."""


@permissions.command("introspect")
@option_db_url
@argument_role
def permissions_introspect(db_url: str, role: str) -> None:
    """Retrieve ACLs from a database."""
    engine = _get_engine(db_url)
    introspect_permissions(engine, role)


@permissions.command("revoke")
@option_db_url
@argument_role
@click.option("-v", "--verbose", count=True)
def permissions_revoke(db_url: str, role: str, verbose: int) -> None:
    """Revoke all table select priviliges for role."""
    engine = _get_engine(db_url)
    revoke_permissions(engine, role, verbose=verbose)


@permissions.command("apply")
@option_db_url
@option_schema_url
@option_profile_url
@click.option(
    "--schema-filename",
    is_flag=False,
    help="Filename of local Amsterdam Schema (single dataset)."
    " If specified, it will be used instead of schema-url",
)
@click.option(
    "--profile-filename",
    is_flag=False,
    help="Filename of local Profile. If specified, it will be used instead of profile-url",
)
@click.option(
    "--pg_schema",
    is_flag=False,
    default="public",
    show_default=True,
    help="Postgres schema containing the data",
)
@click.option(
    "--auto",
    is_flag=True,
    default=False,
    help="Grant each scope X to their associated db role scope_x.",
)
@click.option(
    "--role",
    is_flag=False,
    default="",
    help="Role to receive grants. Ignored when --auto=True",
)
@click.option(
    "--scope",
    is_flag=False,
    default="",
    help="Scope to be granted. Ignored when --auto=True",
)
@click.option(
    "--execute/--dry-run",
    default=False,
    help="Execute SQL statements or dry-run [default]",
)
@click.option(
    "--read/--no-read",
    "set_read_permissions",
    default=True,
    help="Set read permissions [default=read]",
)
@click.option(
    "--write/--no-write",
    "set_write_permissions",
    default=True,
    help="Set dataset-level write permissions [default=write]",
)
@click.option("--create-roles", is_flag=True, default=False, help="Create missing postgres roles")
@click.option(
    "--revoke",
    is_flag=True,
    default=False,
    help="Before granting new permissions, revoke first all previous table and column permissions",
)
@click.option("-v", "--verbose", count=True)
@click.option(
    "-a",
    "--additional-grants",
    multiple=True,
    help="""Additional grants can be defined in the following format:
            `<table_name>:<privilege_1>[,<privilege_n>]*;<grantee_1>[,grantee_n]*`
            Add one option for every table.
            NB: Surround values with double quotes!
              """,
)
def permissions_apply(
    db_url: str,
    schema_url: str,
    profile_url: str,
    schema_filename: str,
    profile_filename: str,
    pg_schema: str,
    auto: bool,
    role: str,
    scope: str,
    execute: bool,
    create_roles: bool,
    set_read_permissions: bool,
    set_write_permissions: bool,
    revoke: bool,
    verbose: int,
    additional_grants: tuple[str] = (),
) -> None:
    """Set permissions for a postgres role.

    This is based on a scope from Amsterdam Schema or Profiles.
    """
    dry_run = not execute

    if auto:
        role = "AUTO"
        scope = "ALL"

    engine = _get_engine(db_url)

    if schema_filename:
        loader = FileSystemSchemaLoader.from_file(schema_filename)
        dataset_schema = loader.get_dataset_from_file(schema_filename)
        ams_schema = {dataset_schema.id: dataset_schema}
    else:
        ams_schema = get_schema_loader(schema_url).get_all_datasets()

    if profile_filename:
        profile = _schema_fetch_url_file(profile_filename)
        profiles = {profile["name"]: profile}
    else:
        # Profiles not live yet, temporarilly commented out
        # profiles = profile_defs_from_url(profiles_url=profile_url)
        profiles = None

    if not (auto or (role and scope)):
        click.echo(
            "Choose --auto or specify both a --role and a --scope to be able to grant permissions"
        )
    elif revoke and not (set_read_permissions and set_write_permissions):
        click.echo(
            "Using --revoke without setting both read and write permissions is destructive."
        )
    else:
        apply_schema_and_profile_permissions(
            engine,
            pg_schema,
            ams_schema,
            profiles,
            role,
            scope,
            set_read_permissions,
            set_write_permissions,
            dry_run,
            create_roles,
            revoke,
            verbose=verbose,
            additional_grants=additional_grants,
        )


def _schema_fetch_url_file(schema_url_file: str) -> dict[str, Any]:
    """Return schemadata from URL or File."""
    # XXX Does not work with datasets that have their tables split
    # out into separate files. Should use _get_dataset_schema instead.

    if not schema_url_file.startswith("http"):
        with open(schema_url_file) as f:
            schema_data = json.load(f)
    else:
        response = requests.get(schema_url_file, timeout=60)
        response.raise_for_status()
        schema_data = response.json()

    return schema_data


@schema.group()
def introspect() -> None:
    """Subcommand to generate a schema."""


@schema.group()
def create() -> None:
    """Subcommand to create a DB object."""


@schema.group()
def diff() -> None:
    """Subcommand to show schema diffs."""


def _fetch_json(location: str) -> dict[str, Any]:
    """Fetch JSON from file or URL.

    Args:
        location: a file name or an URL

    Returns:
        JSON data as a dictionary.
    """
    if not location.startswith("http"):
        schema_file = Path(location)
        if schema_file.is_dir():
            schema_file = schema_file / "schema.json"
        with open(schema_file) as f:
            json_obj = json.load(f)
    else:
        response = requests.get(location, timeout=60)
        response.raise_for_status()
        json_obj = response.json()
    return json_obj


@schema.command()
@option_schema_url
@argument_dataset_id
@click.option(
    "--additional-schemas",
    "-a",
    multiple=True,
    help=(
        "Id of a dataset schema that will be preloaded. "
        "To be used mainly for schemas that are related to the schema that is being validated."
    ),
)
@click.argument("meta_schema_url", nargs=-1)
def validate(
    schema_url: str, dataset_id: str, additional_schemas: list[str], meta_schema_url: tuple[str]
) -> None:
    """Validate a schema against the Amsterdam Schema meta schema.

    Args:

    \b
        DATASET_ID: id of the dataset.
        META_SCHEMA_URL: URL where the meta schema for Amsterdam Schema definitions can be found.
        If multiple are given, schematools will try to validate against the largest version,
        working backwards and stopping at the first version that the objects are valid against.

        Usually META_SCHEMA_URL is something like: https://schemas.data.amsterdam.nl/schema@vn
    """  # noqa: D301,D412,D417
    if not meta_schema_url:
        click.echo("META_SCHEMA_URL not provided", err=True)
        sys.exit(1)

    dataset = _get_dataset_schema(dataset_id, schema_url, prefetch_related=True)

    # The additional schemas are fetched, but the result is not used
    # because the only reason to fetch the additional schemas is to have those schemas
    # available in the cache that is part of the DatasetSchema class
    for schema in additional_schemas:
        _get_dataset_schema(schema, schema_url)

    exit_status = 0

    for meta_schema_version, url in sorted(
        [(version_from_metaschema_url(u), u) for u in set(meta_schema_url)],
        reverse=True,
    ):
        click.echo(f"Validating against metaschema {meta_schema_version}")
        meta_schema = _fetch_json(url)
        if meta_schema_version.major not in COMPATIBLE_METASCHEMAS:
            raise IncompatibleMetaschema(
                f"Schematools {pkg_version} is not compatible"
                f" with metaschema {meta_schema_version}"
            )
        structural_errors = False

        try:
            jsonschema.validate(
                instance=dataset.json_data(inline_tables=True, inline_publishers=False),
                schema=meta_schema,
                format_checker=draft7_format_checker,
            )
        except (jsonschema.ValidationError, jsonschema.SchemaError) as e:
            click.echo("Structural validation: ", nl=False)
            structural_errors = True
            click.echo(format_schema_error(e), err=True)

        semantic_errors = False
        for error in validation.run(dataset):
            if not semantic_errors:  # Only print on first error.
                click.echo("Semantic validation: ", nl=False)
                semantic_errors = True
            click.echo(f"\n{error!s}", err=True)

        if structural_errors or semantic_errors:
            click.echo(f"Dataset is invalid against {meta_schema_version}")
            exit_status = 1
        else:
            click.echo(f"Dataset is valid against {meta_schema_version}")

    sys.exit(exit_status)


def version_from_metaschema_url(url: str) -> SemVer:  # noqa: D103
    return SemVer(url.rpartition("@")[2])


@schema.command()
@option_schema_url
@click.argument("meta_schema_url", nargs=-1)
def validate_publishers(schema_url: str, meta_schema_url: tuple[str]) -> None:
    """Validate all publishers against the Amsterdam Schema meta schema.

    Args:

    \b
        META_SCHEMA_URL: URL where the meta schema for Amsterdam Schema definitions can be found.
        If multiple are given, schematools will try to validate against the largest version,
        working backwards and stopping at the first version that the objects are valid against.

    Options:

    \b
        SCHEMA_URL: URL where the datasets for Amsterdam Schema definitions can be found. The path
        component of this uri is dropped to find the publishers in the root. For example, if
        SCHEMA_URL=https://example.com/datasets, the publishers are extracted from
        https://example.com/publishers.
    """  # noqa: D301,D412,D417
    for meta_schema_version, url in sorted(
        [(version_from_metaschema_url(u), u) for u in set(meta_schema_url)],
        reverse=True,
    ):
        meta_schema = _fetch_json(url)
        if meta_schema_version.major not in COMPATIBLE_METASCHEMAS:
            raise IncompatibleMetaschema(
                f"Schematools {pkg_version} is not"
                f"compatible with metaschema {meta_schema_version}"
            )

        click.echo(f"Validating against metaschema {meta_schema_version}")
        publishers = _get_publishers(schema_url)

        structural_errors = False
        for id_, publisher in publishers.items():
            try:
                click.echo(f"Validating publisher with id {id_}")
                jsonschema.validate(
                    instance=publisher.json_data(),
                    schema=meta_schema,
                    format_checker=draft7_format_checker,
                )
            except (jsonschema.ValidationError, jsonschema.SchemaError) as e:
                click.echo("Structural validation: ", nl=False)
                structural_errors = True
                click.echo(format_schema_error(e), err=True)

        if structural_errors:
            continue
        click.echo(f"All publishers are structurally valid against {meta_schema_version}")
        sys.exit(0)
    click.echo("Publishers are structurally invalid against all supplied metaschema versions")
    sys.exit(1)


@schema.command()
@option_schema_url
@click.argument("meta_schema_url", nargs=-1)
def validate_scopes(schema_url: str, meta_schema_url: tuple[str]) -> None:
    """Validate all scopes against the Amsterdam Schema meta schema.

    Args:

    \b
        META_SCHEMA_URL: URL where the meta schema for Amsterdam Schema definitions can be found.
        If multiple are given, schematools will try to validate against the largest version,
        working backwards and stopping at the first version that the objects are valid against.

    Options:

    \b
        SCHEMA_URL: URL where the datasets for Amsterdam Schema definitions can be found. The path
        component of this uri is dropped to find the scopes in the root. For example, if
        SCHEMA_URL=https://example.com/datasets, the scopes are extracted from
        https://example.com/scopes.
    """  # noqa: D301,D412,D417
    for meta_schema_version, url in sorted(
        [(version_from_metaschema_url(u), u) for u in set(meta_schema_url)],
        reverse=True,
    ):
        meta_schema = _fetch_json(url)
        if meta_schema_version.major not in COMPATIBLE_METASCHEMAS:
            raise IncompatibleMetaschema(
                f"Schematools {pkg_version} is not"
                f"compatible with metaschema {meta_schema_version}"
            )

        click.echo(f"Validating against metaschema {meta_schema_version}")
        scopes = _get_scopes(schema_url)
        structural_errors = False
        for id_, scope in scopes.items():
            try:
                click.echo(f"Validating scope with id {id_}")
                jsonschema.validate(
                    instance=scope.json_data(),
                    schema=meta_schema,
                    format_checker=draft7_format_checker,
                )
            except (jsonschema.ValidationError, jsonschema.SchemaError) as e:
                click.echo("Structural validation: ", nl=False)
                structural_errors = True
                click.echo(format_schema_error(e), err=True)

        if structural_errors:
            continue
        click.echo(f"All scopes are structurally valid against {meta_schema_version}.")
        sys.exit(0)
    click.echo("Scopes are structurally invalid against all supplied metaschema versions")
    sys.exit(1)


@schema.command()
@click.argument("meta_schema_url")
@click.argument("schema_files", nargs=-1)
@click.option(
    "-m",
    "--extra_meta_schema_url",
    help="An additional metaschema to try validation in case meta_schema_url fails",
)
def batch_validate(
    meta_schema_url: str, schema_files: tuple[str], extra_meta_schema_url: str
) -> None:
    """Batch validate schemas.

    This command was tailored so that it could be run from a pre-commit hook.
    The order and type of its arguments differ from other `schema` sub-commands:
    the files it accepts are either "dataset.json" files or files denoting
    tables. In the case of table files, the containing dataset will be
    validated.

    It will perform both structural and semantic validation of schemas.
    If extra_meta_schema_url is supplied, meta_schema_url will be tried first.

    Args:

    \b
        META_SCHEMA_URL: the URL to the Amsterdam meta schema
        SCHEMA_FILES: one or more schema files to be validated
    """  # noqa: D301,D412,D417
    errors: defaultdict[str, defaultdict[str, list[str]]] = defaultdict(lambda: defaultdict(list))

    # Find the root "datasets" directory.
    datasets_dir = Path(os.path.commonpath(schema_files)).absolute().resolve()
    path_parts = datasets_dir.parts

    # Bail out if there is no `datasets` directory
    try:
        datasets_idx = path_parts.index("datasets")
    except ValueError:
        raise ValueError("dataset files do not live in a common 'datasets' dir") from None

    # Find out if we need to go up the directory tree to get at `datasets` dir
    # This could be needed if we are only checking one dataset,
    # because in that case datasets_dir is initially the full path to
    # the `dataset.json` file.
    up_tree_count = len(path_parts) - 2 - datasets_idx
    if up_tree_count > 0:
        datasets_dir = datasets_dir.parents[up_tree_count]

    loader = FileSystemSchemaLoader(datasets_dir)

    done = set()
    for schema_file in schema_files:
        # If the schema file is a table, find the dataset.json
        # file in one of the parent directories.
        ds_dir = Path(schema_file).parent
        while not (ds_dir / "dataset.json").exists():
            ds_dir = ds_dir.parent

        # Don't run validations multiple times when several files in a dataset have changed.
        main_file = os.path.join(ds_dir, "dataset.json")
        if main_file in done:
            continue

        meta_schema_urls = [meta_schema_url]
        if extra_meta_schema_url:
            meta_schema_urls.append(extra_meta_schema_url)
        for url in meta_schema_urls:
            meta_schema_version = version_from_metaschema_url(url)
            click.echo(f"Validating {main_file} against {meta_schema_version}")

            try:
                dataset = loader.get_dataset_from_file(main_file)
            except ValueError as ve:
                errors[schema_file][meta_schema_version].append(str(ve))
                # No sense in continuing if we can't read the schema file.
                break

            meta_schema = _fetch_json(url)
            try:
                jsonschema.validate(
                    instance=dataset.json_data(inline_tables=True, inline_publishers=False),
                    schema=meta_schema,
                    format_checker=draft7_format_checker,
                )
            except (jsonschema.ValidationError, jsonschema.SchemaError) as struct_error:
                errors[schema_file][meta_schema_version].append(format_schema_error(struct_error))

            for sem_error in validation.run(dataset, main_file):
                errors[schema_file][meta_schema_version].append(str(sem_error))

            if not errors[schema_file][meta_schema_version]:
                click.echo(f"{schema_file} is valid against meta schema {meta_schema_version}")
                # We dont show errors if the file is valid against one of the metaschemas
                errors.pop(schema_file)
                break
        done.add(main_file)

    if errors:
        width = len(max(errors.keys(), key=lambda x: len(x)))
        for schema_file, versions in errors.items():
            click.echo(f"{schema_file} is invalid against all metaschema versions")
            version_width = len(max(versions.keys(), key=lambda x: len(x)))
            for version_, err_msgs in versions.items():
                for msg in err_msgs:
                    click.echo(f"{schema_file:>{width}} - {version_:>{version_width}}: {msg}")
        sys.exit(1)


def format_schema_error(e: jsonschema.SchemaError | jsonschema.ValidationError) -> str:
    s = io.StringIO()
    s.write(f"{e.json_path}, {list(e.schema_path)}")
    if e.message and not e.context:
        s.write("\n\t" + e.message)
    if e.context:
        for ec in e.context:
            s.write("\n\t" + ec.message.strip())
    return s.getvalue()


@schema.command("ckan")
@option_schema_url
@click.option(
    "--upload-url",
    "-u",
    default=None,
    help="URL for uploading, e.g., https://data.overheid.nl/data/ (none to print to stdout)",
)
def to_ckan(schema_url: str, upload_url: str):
    """Convert all schemas to CKAN format, and optionally upload them.

    The API key for CKAN is taken from the environment variable CKAN_API_KEY.
    """
    api_key = os.getenv("CKAN_API_KEY")
    if upload_url is not None and api_key is None:
        click.echo("CKAN_API_KEY not set in environment", err=True)
        exit(1)

    status = 0

    datasets = get_schema_loader(schema_url).get_all_datasets()

    data = []
    for path, ds in datasets.items():
        if ds.status != DatasetSchema.Status.beschikbaar:
            continue
        try:
            data.append(ckan.from_dataset(ds, path))
        except Exception as e:  # noqa: BLE001
            logger.error("in dataset %s: %s", ds.identifier, str(e))  # noqa: G200
            status = 1

    if upload_url is None:
        for ds in data:
            click.echo(ds)
        exit(0)

    headers = {"Authorization": api_key}
    for ds in data:
        ident = ds["identifier"]
        url = f"{upload_url}/api/3/action/package_update?id={ident}"
        response = requests.post(url, headers=headers, json=ds, timeout=60)
        logger.debug("%s: %d, %s", url, response.status_code, response.json())

        if response.status_code == 404:
            # 404 *can* mean no such dataset. Try again with package_create.
            url = upload_url + "/api/3/action/package_create"
            response = requests.post(url, headers=headers, json=ds, timeout=60)
            logger.debug("%s: %d, %s", url, response.status_code, response.json())

        if not (200 <= response.status_code < 300):
            logger.error("uploading %s: %s", ident, response.json())
            status = 1

    exit(status)


@show.command("provenance")
@click.argument("dataset_id")
def show_provenance(dataset_id: str, schema_url: str) -> None:
    """Retrieve the key-values pairs of the source column.

    (specified as a 'provenance' property of an attribute)
    and its translated name (the attribute name itself)
    """
    dataset = _get_dataset_schema(dataset_id, schema_url, prefetch_related=True)
    try:
        instance = ProvenanceIteration(dataset)
        click.echo(instance.final_dic)
    except (jsonschema.ValidationError, jsonschema.SchemaError, KeyError) as e:
        click.echo(str(e), err=True)
        exit(1)


@show.command("tablenames")
@option_db_url
def show_tablenames(db_url: str) -> None:
    """Retrieve tablenames from a database."""
    engine = _get_engine(db_url)
    names = inspect(engine).get_table_names()
    click.echo("\n".join(names))


@show.command("datasets")
@option_schema_url
@click.option("--to-snake-case", "snake_it", is_flag=True)
def show_datasets(schema_url: str, snake_it: bool) -> None:
    """Retrieve the ids of all the datasets."""
    loader = get_schema_loader(schema_url)
    modifier = to_snake_case if snake_it else lambda x: x
    for dataset_schema in loader.get_all_datasets().values():
        click.echo(modifier(dataset_schema.id))


@show.command("datasettables")
@option_schema_url
@argument_dataset_id
@click.option("--to-snake-case", "snake_it", is_flag=True)
def show_datasettables(schema_url: str, dataset_id: str, snake_it: bool) -> None:
    """Retrieve the ids of the datasettables for the indicated dataset."""
    dataset_schema = _get_dataset_schema(dataset_id, schema_url, prefetch_related=False)
    modifier = to_snake_case if snake_it else lambda x: x
    for dataset_table in dataset_schema.tables:
        click.echo(modifier(dataset_table.id))


@show.command("mapfile")
@click.option(
    "--schema-url",
    envvar="SCHEMA_URL",
    default=DEFAULT_SCHEMA_URL,
    show_default=True,
    required=True,
    help="Url where valid amsterdam schema files are found. "
    "SCHEMA_URL can also be provided as environment variable.",
)
@click.argument("dataset_id")
def show_mapfile(schema_url: str, dataset_id: str) -> None:
    """Generate a mapfile based on a dataset schema."""
    try:
        dataset_schema = get_schema_loader(schema_url).get_dataset(dataset_id)
    except KeyError:
        raise click.BadParameter(f"Schema {dataset_id} not found.") from None
    click.echo(create_mapfile(dataset_schema))


@show.command("scopes")
@option_schema_url
@argument_dataset_id
def show_scopes(schema_url: str, dataset_id: str) -> None:
    """Generate a list of all the scopes used in the indicated dataset."""
    # We need to `unfreeze` the sets, to make the set operations work.
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    scopes = set(dataset_schema.auth)
    for table in dataset_schema.tables:
        scopes |= set(table.auth) | reduce(operator.or_, (set(f.auth) for f in table.fields))
    click.echo(" ".join(str(scope) for scope in scopes - {"OPENBAAR"}))


@introspect.command("db")
@click.option("--db-schema", "-s", help="Tables are in a different postgres schema (not 'public')")
@click.option("--prefix", "-p", help="Tables have prefix that needs to be stripped")
@option_db_url
@click.argument("dataset_id")
@click.argument("tables", nargs=-1)
def introspect_db(
    db_schema: str, prefix: str, db_url: str, dataset_id: str, tables: Iterable[str]
) -> None:
    """Generate a schema for the tables in a database."""
    engine = _get_engine(db_url)
    aschema = introspect_db_schema(engine, dataset_id, tables, db_schema, prefix)
    click.echo(json.dumps(aschema, indent=2))


@introspect.command("geojson")
@click.argument("dataset_id")
@click.argument("files", nargs=-1, required=True)
def introspect_geojson(dataset_id: str, files: list[str]) -> None:
    """Generate a schema from a GeoJSON file."""
    aschema = introspect_geojson_files(dataset_id, files)
    click.echo(json.dumps(aschema, indent=2))


@import_.command("ndjson")
@option_db_url
@option_schema_url
@argument_dataset_id
@click.argument("table_id")
@click.argument("ndjson_path")
@click.option("--batch_size", default=100, type=int)
@click.option("--truncate-table", is_flag=True)
def import_ndjson(
    db_url: str,
    schema_url: str,
    dataset_id: str,
    table_id: str,
    batch_size: int,
    ndjson_path: PosixPath,
    truncate_table: bool,
) -> None:
    """Import a NDJSON file into a table."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    importer = NDJSONImporter(dataset_schema, engine)
    importer.generate_db_objects(table_id, truncate=False, ind_tables=True, ind_extra_index=False)
    importer.load_file(ndjson_path, batch_size, truncate=truncate_table)


@import_.command("geojson")
@option_db_url
@option_schema_url
@argument_dataset_id
@click.argument("table_id")
@click.argument("geojson_path")
@click.option("--batch_size", default=100, type=int)
@click.option("--truncate-table", is_flag=True)
def import_geojson(
    db_url: str,
    schema_url: str,
    dataset_id: str,
    table_id: str,
    batch_size: int,
    geojson_path: PosixPath,
    truncate_table: bool,
) -> None:
    """Import a GeoJSON file into a table."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    importer = GeoJSONImporter(dataset_schema, engine)
    importer.generate_db_objects(table_id, truncate=False, ind_tables=True, ind_extra_index=False)
    importer.load_file(geojson_path, batch_size=batch_size, truncate=truncate_table)


def _get_dataset_schema(
    dataset_id: str, schema_url: str, prefetch_related: bool = False
) -> DatasetSchema:
    """Find the dataset schema for the given dataset.

    Args:
        dataset_id: id of the dataset.
        schema_url: url of the location where the collection of amsterdam schemas is found.
        prefetch_related: related schemas should be prefetched.
    """
    loader = get_schema_loader(schema_url)
    try:
        return loader.get_dataset(dataset_id, prefetch_related=prefetch_related)
    except DatasetNotFound as e:
        raise click.ClickException(str(e)) from None


def _get_publishers(schema_url: str) -> dict[str, Publisher]:
    """Find the publishers from the given schema_url.

    Args:
        schema_url: url of the location where the collection of amsterdam schemas is found.
    """
    loader = get_schema_loader(schema_url)
    try:
        return loader.get_all_publishers()
    except SchemaObjectNotFound as e:
        raise click.ClickException(str(e)) from None


def _get_scopes(schema_url: str) -> dict[str, Scope]:
    """Find the scopes from the given schema_url.

    Args:
        schema_url: url of the location where the collection of amsterdam schemas is found.
    """
    loader = get_schema_loader(schema_url)
    try:
        return loader.get_all_scopes()
    except (SchemaObjectNotFound, DuplicateScopeId) as e:
        raise click.ClickException(str(e)) from None


@create.command("extra_index")
@option_db_url
@option_schema_url
@argument_dataset_id
def create_identifier_index(db_url: str, schema_url: str, dataset_id: str) -> None:
    """Execute SQLalchemy Index based on Identifier in the JSON schema data definition."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    importer = BaseImporter(dataset_schema, engine)

    for table in dataset_schema.get_tables():
        importer.generate_db_objects(
            table.id,
            ind_tables=False,
            ind_extra_index=True,
            is_versioned_dataset=importer.is_versioned_dataset,
        )


@create.command("tables")
@option_db_url
@option_schema_url
@argument_dataset_id
def create_tables(db_url: str, schema_url: str, dataset_id: str) -> None:
    """Execute SQLalchemy Table objects."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url, prefetch_related=True)
    importer = BaseImporter(dataset_schema, engine)

    for table in dataset_schema.get_tables():
        importer.generate_db_objects(
            table.id,
            ind_extra_index=False,
            ind_tables=True,
            is_versioned_dataset=importer.is_versioned_dataset,
        )


@create.command("sql")
@click.option("--versioned/--no-versioned", default=True)
@option_db_url
@click.argument("schema_path")
def create_sql(versioned: bool, db_url: str, schema_path: str) -> None:
    """Generate SQL Create from amsterdam schema definition."""
    engine = _get_engine(db_url)
    loader = FileSystemSchemaLoader.from_file(schema_path)
    dataset_schema = loader.get_dataset_from_file(schema_path)
    tables = tables_factory(dataset_schema, is_versioned_dataset=versioned)
    for table in tables.values():
        table_sql = CreateTable(table).compile(engine)
        click.echo(str(table_sql))


@create.command("all")
@option_db_url
@option_schema_url
@click.option(
    "-x",
    "--exclude",
    multiple=True,
    type=str,
    default=[],
    help="dataset_id to exclude. Can be repeated.",
)
@click.argument("dataset_id", required=False)
def create_all_objects(
    db_url: str, schema_url: str, exclude: list[str], dataset_id: str | None
) -> None:
    """Execute SQLalchemy Index (Identifier fields) and Table objects.

    If no DATASET_ID is provide it will process all datasets!
    """
    loader = get_schema_loader(schema_url)
    if dataset_id is None:
        click.echo("No 'dataset_id' provided. Processing all datasets!")
        dataset_schemas = loader.get_all_datasets().values()
    else:
        dataset_schemas = [loader.get_dataset(dataset_id, prefetch_related=True)]

    engine = _get_engine(db_url)
    for dataset_schema in dataset_schemas:
        if dataset_schema.id in exclude:
            msg = f"Skipping dataset {dataset_id!r}"
            click.echo(msg)
            click.echo("=" * len(msg))
            continue
        msg = f"Processing dataset {dataset_id!r}"
        click.echo(msg)
        click.echo("-" * len(msg))
        importer = BaseImporter(dataset_schema, engine)
        for table in dataset_schema.get_tables():
            importer.generate_db_objects(
                table.id,
                is_versioned_dataset=importer.is_versioned_dataset,
            )


@diff.command("all")
@option_schema_url
@click.argument("diff_schema_url")
def diff_schemas(schema_url: str, diff_schema_url: str) -> None:
    """Show diff for two sets of schemas.

    The left-side schemas location is
    defined in SCHEMA_URL (or via --schema-url), the right-side schemas location
    has to be on the command-line.

    This can be used to compare two sets of schemas, e.g. ACC and PRD schemas.

    For nicer output, pipe it through a json formatter.
    """
    schemas = get_schema_loader(schema_url).get_all_datasets()
    diff_schemas = get_schema_loader(diff_schema_url).get_all_datasets()
    click.echo(DeepDiff(schemas, diff_schemas, ignore_order=True).to_json())


@export.command("geopackage")
@option_db_url
@option_schema_url
@argument_dataset_id
@click.option("--output", "-o", default="/tmp")  # noqa: S108  # nosec: B108
@click.option("--table-ids", "-t", multiple=True)
@click.option(
    "--scopes",
    "-s",
    multiple=True,
    help="Scopes option has been disabled for now, only public data can be exported.",
)
@click.option("--size")
def export_geopackages_for(
    db_url: str,
    schema_url: str,
    dataset_id: str,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
) -> None:
    """Export geopackages from postgres."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    with engine.begin() as connection:
        export_geopackages(
            connection, dataset_schema, output, table_ids=table_ids, scopes=[], size=size
        )


@export.command("csv")
@option_db_url
@option_schema_url
@argument_dataset_id
@click.option("--output", "-o", default="/tmp")  # noqa: S108  # nosec: B108
@click.option("--table-ids", "-t", multiple=True)
@click.option(
    "--scopes",
    "-s",
    multiple=True,
    help="Scopes option has been disabled for now, only public data can be exported.",
)
@click.option("--size", type=int)
def export_csvs_for(
    db_url: str,
    schema_url: str,
    dataset_id: str,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
) -> None:
    """Export csv files from postgres."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    with engine.begin() as connection:
        export_csvs(connection, dataset_schema, output, table_ids=table_ids, scopes=[], size=size)


@export.command("jsonlines")
@option_db_url
@option_schema_url
@argument_dataset_id
@click.option("--output", "-o", default="/tmp")  # noqa: S108  # nosec: B108
@click.option("--size")
@click.option("--table-ids", "-t", multiple=True)
@click.option(
    "--scopes",
    "-s",
    multiple=True,
    help="Scopes option has been disabled for now, only public data can be exported.",
)
@click.option("--size")
def export_jsonls_for(
    db_url: str,
    schema_url: str,
    dataset_id: str,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
) -> None:
    """Export csv files from postgres."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(dataset_id, schema_url)
    with engine.begin() as connection:
        export_jsonls(
            connection, dataset_schema, output, table_ids=table_ids, scopes=[], size=size
        )


@tocase.command("camel")
@click.argument("input_str")
def convert_to_camel_case(input_str: str) -> str:
    """Converts INPUT_STR to camel case."""
    click.echo(toCamelCase(input_str))


@tocase.command("snake")
@click.argument("input_str")
def convert_to_snake_case(input_str: str) -> str:
    """Converts INPUT_STR to snake case."""
    click.echo(to_snake_case(input_str))


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import logging

import pytest
from psycopg2.errors import DuplicateObject
from sqlalchemy.exc import ProgrammingError

from schematools.importer.ndjson import NDJSONImporter
from schematools.permissions.db import apply_schema_and_profile_permissions
from schematools.types import DatasetSchema, Scope


# In test files we use a lot of non-existent scopes, so instead of writing scope
# json files we monkeypatch this method.
@pytest.fixture(autouse=True)
def patch_find_scope_by_id(monkeypatch):
    monkeypatch.setattr(DatasetSchema, "_find_scope_by_id", Scope.from_string)


class TestReadPermissions:
    def test_auto_permissions(self, here, engine, gebieden_schema_auth, dbsession):
        """
        Prove that roles are automatically created for each scope in the schema
        LEVEL/A --> scope_level_a
        LEVEL/B --> scope_level_b
        LEVEL/C --> scope_level_c
        """
        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_auth, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {gebieden_schema_auth.id: gebieden_schema_auth}
        profile_path = here / "files" / "profiles" / "gebieden_test.json"
        with open(profile_path) as f:
            profile = json.load(f)
        profiles = {profile["name"]: profile}

        # Apply the permissions from Schema and Profiles.
        apply_schema_and_profile_permissions(
            engine, "public", ams_schema, profiles, "AUTO", "ALL", create_roles=True
        )
        _check_select_permission_granted(engine, "scope_level_a", "gebieden_buurten_v1")
        _check_select_permission_granted(
            engine, "scope_level_b", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "scope_level_c", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )

    def test_auto_permissions_with_scopes(self, here, engine, gebieden_schema_scopes, dbsession):
        """
        Prove that roles are automatically created for each Scope objects.

        Scope objects are properly resolved and permissions granted.
        Using Scope objects in the scopes/HARRY folder on dataset, table, and field level.
        Ensure we can still use auth strings in the same schema.
        """
        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_scopes, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {gebieden_schema_scopes.id: gebieden_schema_scopes}
        profile_path = here / "files" / "profiles" / "gebieden_test.json"
        with open(profile_path) as f:
            profile = json.load(f)
        profiles = {profile["name"]: profile}

        # Apply the permissions from Schema and Profiles.
        apply_schema_and_profile_permissions(
            engine, "public", ams_schema, profiles, "AUTO", "ALL", create_roles=True
        )
        _check_select_permission_granted(engine, "scope_harry_one", "gebieden_buurten_v1")
        _check_select_permission_granted(
            engine, "scope_harry_two", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "scope_harry_three", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )

        # Check if auth strings in the same schema also work.
        _check_select_permission_granted(
            engine,
            "scope_level_d",
            "gebieden_bouwblokken_v1",
            "ligt_in_buurt_id, ligt_in_buurt_loose_id",
        )

    def test_nm_relations_permissions(
        self, here, engine, kadastraleobjecten_schema, dbsession, caplog
    ):
        importer = NDJSONImporter(kadastraleobjecten_schema, engine)
        importer.generate_db_objects("kadastraleobjecten", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        # This schema has auth on dataset level, not on table
        ams_schema = {kadastraleobjecten_schema.id: kadastraleobjecten_schema}

        _create_role(engine, "openbaar")
        _create_role(engine, "brk_rsn")
        _create_role(engine, "brk_ro")

        # Check if the roles exist, the tables exist,
        # and the roles have no read privilege on the tables.
        _check_select_permission_denied(engine, "openbaar", "brk_kadastraleobjecten_v1")
        _check_select_permission_denied(engine, "brk_rsn", "brk_kadastraleobjecten_v1")
        _check_select_permission_denied(engine, "brk_ro", "brk_kadastraleobjecten_v1")

        # make sure role 'write_brk' exists with create_roles=True
        # The role exists now for all test following this statement
        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine,
                "public",
                ams_schema,
                {},
                "openbaar",
                "OPENBAAR",
                create_roles=True,
                verbose=1,
            )
            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, {}, "brk_rsn", "BRK/RSN", verbose=1
            )
            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, {}, "brk_ro", "BRK/RO", verbose=1
            )

        grants = _filter_grant_statements(caplog)
        assert grants == [
            "GRANT SELECT (begin_geldigheid) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (eind_geldigheid) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (id) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (identificatie) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (koopsom) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_ro",
            "GRANT SELECT (neuron_id) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (registratiedatum) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (soort_cultuur_onbebouwd_code) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_ro",
            "GRANT SELECT (soort_cultuur_onbebouwd_omschrijving) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_ro",
            "GRANT SELECT (soort_grootte) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT (volgnummer) ON TABLE public.brk_kadastraleobjecten_v1 TO brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject__id_seq TO brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1 TO brk_rsn",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_v1 TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject__id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject__id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject__id_seq TO write_brk",
        ]

        # table denied
        _check_select_permission_denied(engine, "openbaar", "brk_kadastraleobjecten_v1")
        _check_select_permission_denied(engine, "openbaar", "brk_kadastraleobjecten_v1", "koopsom")

        # table denied, column granted, auth level dataset
        _check_select_permission_denied(engine, "brk_rsn", "brk_kadastraleobjecten_v1")
        _check_select_permission_denied(engine, "brk_rsn", "brk_kadastraleobjecten_v1", "koopsom")
        _check_select_permission_granted(
            engine, "brk_rsn", "brk_kadastraleobjecten_v1", "identificatie"
        )

        # table denied, column granted, auth level field
        _check_select_permission_denied(engine, "brk_ro", "brk_kadastraleobjecten_v1")
        _check_select_permission_granted(engine, "brk_ro", "brk_kadastraleobjecten_v1", "koopsom")

        # nm relations table tests, should have dataset auth level: brk_rsn
        _check_select_permission_denied(
            engine, "openbaar", "brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1"
        )
        _check_select_permission_denied(
            engine, "brk_ro", "brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1"
        )
        _check_select_permission_granted(
            engine, "brk_rsn", "brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1"
        )

    def test_brk_permissions(
        self, here, engine, brk_schema_without_bag_relations, dbsession, caplog
    ):
        """Prove that a dataset with many nested tables get the proper permissions."""
        importer = NDJSONImporter(brk_schema_without_bag_relations, engine)
        for table in brk_schema_without_bag_relations.get_tables():
            importer.generate_db_objects(table.id, truncate=True, ind_extra_index=False)

        # Setup schema and profile
        # This schema has auth on dataset level, not on table
        ams_schema = {brk_schema_without_bag_relations.id: brk_schema_without_bag_relations}

        # make sure role 'write_brk' exists with create_roles=True
        # The role exists now for all test following this statement
        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine,
                "public",
                ams_schema,
                {},
                "AUTO",
                "ALL",
                create_roles=True,
                verbose=1,
            )

        grants = _filter_grant_statements(caplog)
        assert grants == [
            "GRANT SELECT ON SEQUENCE public.brk_aantekeningenkadastraleobjecten_heeft_betrokken_pers_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_aantekeningenrechten_heeft_betrokken_persoon_v1_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_aantekeningenrechten_is_gbsd_op_sdl_v1_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_kadastraleobjecten_hft_rel_mt_vot_v1_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_kadastraleobjecten_soort_cultuur_bebouwd_v1_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_stukdelen_is_bron_voor_aantekening_kadastraal_object_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_stukdelen_is_bron_voor_aantekening_recht_v1_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON SEQUENCE public.brk_stukdelen_is_bron_voor_zakelijk_recht_v1_id_seq TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_aantekeningenkadastraleobjecten_heeft_betrokken_persoon_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_aantekeningenkadastraleobjecten_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_aantekeningenrechten_heeft_betrokken_persoon_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_aantekeningenrechten_is_gbsd_op_sdl_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_aantekeningenrechten_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_aardzakelijkerechten_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_gemeentes_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastralegemeentecodes_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastralegemeentes_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastraleobjecten_hft_rel_mt_vot_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastraleobjecten_soort_cultuur_bebouwd_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastraleobjecten_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastralesecties_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_kadastralesubjecten_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_meta_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_stukdelen_is_bron_voor_aantekening_kadastraal_object_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_stukdelen_is_bron_voor_aantekening_recht_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_stukdelen_is_bron_voor_zakelijk_recht_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_stukdelen_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_tenaamstellingen_v1 TO scope_brk_rsn",
            "GRANT SELECT ON TABLE public.brk_zakelijkerechten_v1 TO scope_brk_rsn",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_aantekeningenkadastraleobjecten_heeft_betrokken_persoon_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_aantekeningenkadastraleobjecten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_aantekeningenrechten_heeft_betrokken_persoon_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_aantekeningenrechten_is_gbsd_op_sdl_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_aantekeningenrechten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_aardzakelijkerechten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_gemeentes_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastralegemeentecodes_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastralegemeentes_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_hft_rel_mt_vot_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_soort_cultuur_bebouwd_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastraleobjecten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastralesecties_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_kadastralesubjecten_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_meta_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_stukdelen_is_bron_voor_aantekening_kadastraal_object_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_stukdelen_is_bron_voor_aantekening_recht_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_stukdelen_is_bron_voor_zakelijk_recht_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_stukdelen_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_tenaamstellingen_v1 TO write_brk",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.brk_zakelijkerechten_v1 TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_aantekeningenkadastraleobjecten_heeft_betrokken_pers_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_aantekeningenrechten_heeft_betrokken_persoon_v1_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_aantekeningenrechten_is_gbsd_op_sdl_v1_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_kadastraleobjecten_hft_rel_mt_vot_v1_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_kadastraleobjecten_soort_cultuur_bebouwd_v1_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_stukdelen_is_bron_voor_aantekening_kadastraal_object_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_stukdelen_is_bron_voor_aantekening_recht_v1_id_seq TO write_brk",
            "GRANT USAGE ON SEQUENCE public.brk_stukdelen_is_bron_voor_zakelijk_recht_v1_id_seq TO write_brk",
        ]

    def test_openbaar_permissions(self, here, engine, afval_schema, dbsession, caplog):
        """
        Prove that the default auth scope is "OPENBAAR".
        """

        importer = NDJSONImporter(afval_schema, engine)
        importer.generate_db_objects("containers", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("clusters", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {afval_schema.id: afval_schema}
        profile_path = here / "files" / "profiles" / "gebieden_test.json"
        with open(profile_path) as f:
            profile = json.load(f)
        profiles = {profile["name"]: profile}

        # Create postgres roles
        _create_role(engine, "openbaar")
        _create_role(engine, "bag_r")
        # Check if the roles exist, the tables exist,
        # and the roles have no read privilige on the tables.
        _check_select_permission_denied(engine, "openbaar", "afvalwegingen_containers_v1")
        _check_select_permission_denied(engine, "bag_r", "afvalwegingen_clusters_v1")

        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine=engine,
                pg_schema="public",
                ams_schema=ams_schema,
                profiles=profiles,
                role="openbaar",
                scope="OPENBAAR",
                create_roles=True,
            )
            apply_schema_and_profile_permissions(
                engine=engine,
                pg_schema="public",
                ams_schema=ams_schema,
                profiles=profiles,
                role="bag_r",
                scope="BAG/R",
                create_roles=True,
                verbose=1,
            )

        grants = _filter_grant_statements(caplog)
        assert grants == [
            "GRANT SELECT ON TABLE public.afvalwegingen_clusters_v1 TO bag_r",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.afvalwegingen_clusters_v1 TO write_afvalwegingen",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.afvalwegingen_containers_v1 TO write_afvalwegingen",
        ]

        _check_select_permission_granted(engine, "openbaar", "afvalwegingen_containers_v1")
        _check_select_permission_denied(engine, "openbaar", "afvalwegingen_clusters_v1")
        _check_select_permission_denied(engine, "bag_r", "afvalwegingen_containers_v1")
        _check_select_permission_granted(engine, "bag_r", "afvalwegingen_clusters_v1")

    def test_interacting_permissions(self, here, engine, gebieden_schema_auth, dbsession):
        """
        Prove that dataset, table, and field permissions are set
        according to the "OF-OF" Exclusief principle:

        * Een user met scope LEVEL/A mag alles uit de dataset gebieden zien,
          behalve tabel bouwblokken.
        * Een user met scope LEVEL/B mag alle velden van tabel bouwblokken zien,
          behalve beginGeldigheid.
        * Een user met scope LEVEL/C mag veld beginGeldigheid zien.
        """

        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_auth, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {gebieden_schema_auth.id: gebieden_schema_auth}
        profile_path = here / "files" / "profiles" / "gebieden_test.json"
        with open(profile_path) as f:
            profile = json.load(f)
        profiles = {profile["name"]: profile}

        # Create postgres roles
        test_roles = ["level_a", "level_b", "level_c"]
        for test_role in test_roles:
            _create_role(engine, test_role)

        # Check if the roles exist, the tables exist,
        # and the roles have no read privilige on the tables.
        for test_role in test_roles:
            for table in ["gebieden_bouwblokken_v1", "gebieden_buurten_v1"]:
                _check_select_permission_denied(engine, test_role, table)

        # Apply the permissions from Schema and Profiles.
        apply_schema_and_profile_permissions(
            engine, "public", ams_schema, profiles, "level_a", "LEVEL/A"
        )
        apply_schema_and_profile_permissions(
            engine, "public", ams_schema, profiles, "level_b", "LEVEL/B"
        )
        apply_schema_and_profile_permissions(
            engine, "public", ams_schema, profiles, "level_c", "LEVEL/C"
        )

        # Check if the read priviliges are correct
        _check_select_permission_denied(engine, "level_a", "gebieden_bouwblokken_v1")
        _check_select_permission_granted(engine, "level_a", "gebieden_buurten_v1")

        _check_select_permission_granted(
            engine, "level_b", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_denied(
            engine, "level_b", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "level_b", "gebieden_buurten_v1")

        _check_select_permission_denied(
            engine, "level_c", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "level_c", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "level_c", "gebieden_buurten_v1")

    def test_auth_list_permissions(
        self, here, engine, gebieden_schema_auth_list, dbsession, caplog
    ):
        """
        Prove that dataset, table, and field permissions are set,
        according to the "OF-OF" Exclusief principle.
        Prove that when the auth property is a list of scopes, this is interpreted as "OF-OF".

        * Een user met scope LEVEL/A1 of LEVEL/A2 mag alles uit de dataset gebieden zien,
          behalve tabel bouwblokken.
        * Een user met scope LEVEL/B1 of LEVEL/B2 mag alle velden van tabel bouwblokken zien,
          behalve beginGeldigheid.
        * Een user met scope LEVEL/C1 of LEVEL/B2 mag veld beginGeldigheid zien.
        """

        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_auth_list, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {gebieden_schema_auth_list.id: gebieden_schema_auth_list}
        profile_path = here / "files" / "profiles" / "gebieden_test.json"
        with open(profile_path) as f:
            profile = json.load(f)
        profiles = {profile["name"]: profile}

        # Create postgres roles
        test_roles = [
            "level_a1",
            "level_a2",
            "level_b1",
            "level_b2",
            "level_c1",
            "level_c2",
        ]
        for test_role in test_roles:
            _create_role(engine, test_role)

        # Check if the roles exist, the tables exist,
        # and the roles have no read privilige on the tables.
        for test_role in test_roles:
            for table in ["gebieden_bouwblokken_v1", "gebieden_buurten_v1"]:
                _check_select_permission_denied(engine, test_role, table)

        # Apply the permissions from Schema and Profiles.
        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "level_a1", "LEVEL/A1", verbose=1
            )
            grants = _filter_grant_statements(caplog)
            assert grants == [
                "GRANT SELECT ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO level_a1",
                "GRANT SELECT ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO level_a1",
                "GRANT SELECT ON TABLE public.gebieden_buurten_v1 TO level_a1",
                "GRANT SELECT ON TABLE public.gebieden_wijken_v1 TO level_a1",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
                "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            ]

            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "level_b1", "LEVEL/B1", verbose=1
            )
            grants = _filter_grant_statements(caplog)
            assert grants == [
                "GRANT SELECT (eind_geldigheid) ON TABLE public.gebieden_bouwblokken_v1 TO level_b1",
                "GRANT SELECT (id) ON TABLE public.gebieden_bouwblokken_v1 TO level_b1",
                "GRANT SELECT (ligt_in_buurt_id) ON TABLE public.gebieden_bouwblokken_v1 TO level_b1",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
                "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            ]

            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "level_c1", "LEVEL/C1", verbose=1
            )
            grants = _filter_grant_statements(caplog)
            assert grants == [
                "GRANT SELECT (begin_geldigheid) ON TABLE public.gebieden_bouwblokken_v1 TO level_c1",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
                "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            ]

            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "level_a2", "LEVEL/A2", verbose=1
            )
            grants = _filter_grant_statements(caplog)
            assert grants == [
                "GRANT SELECT ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO level_a2",
                "GRANT SELECT ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO level_a2",
                "GRANT SELECT ON TABLE public.gebieden_buurten_v1 TO level_a2",
                "GRANT SELECT ON TABLE public.gebieden_wijken_v1 TO level_a2",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
                "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            ]

            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "level_b2", "LEVEL/B2", verbose=1
            )
            grants = _filter_grant_statements(caplog)
            assert grants == [
                "GRANT SELECT (eind_geldigheid) ON TABLE public.gebieden_bouwblokken_v1 TO level_b2",
                "GRANT SELECT (id) ON TABLE public.gebieden_bouwblokken_v1 TO level_b2",
                "GRANT SELECT (ligt_in_buurt_id) ON TABLE public.gebieden_bouwblokken_v1 TO level_b2",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
                "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            ]

            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "level_c2", "LEVEL/C2", verbose=1
            )
            grants = _filter_grant_statements(caplog)
            assert grants == [
                "GRANT SELECT (begin_geldigheid) ON TABLE public.gebieden_bouwblokken_v1 TO level_c2",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
                "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
                "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            ]

        # Check if the read priviliges are correct
        _check_select_permission_denied(engine, "level_a1", "gebieden_bouwblokken_v1")
        _check_select_permission_granted(engine, "level_a1", "gebieden_buurten_v1")
        _check_select_permission_denied(engine, "level_a2", "gebieden_bouwblokken_v1")
        _check_select_permission_granted(engine, "level_a2", "gebieden_buurten_v1")

        _check_select_permission_granted(
            engine, "level_b1", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_denied(
            engine, "level_b1", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "level_b1", "gebieden_buurten_v1")
        _check_select_permission_granted(
            engine, "level_b2", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_denied(
            engine, "level_b2", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "level_b2", "gebieden_buurten_v1")

        _check_select_permission_denied(
            engine, "level_c1", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "level_c1", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "level_c1", "gebieden_buurten_v1")

        _check_select_permission_denied(
            engine, "level_c2", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "level_c2", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "level_c2", "gebieden_buurten_v1")

        # Check that there are no INSERT, UPDATE, TRUNCATE, DELETE privileges
        _check_insert_permission_denied(
            engine, "level_b1", "gebieden_bouwblokken_v1", "id", "'abc'"
        )
        _check_update_permission_denied(
            engine, "level_b1", "gebieden_bouwblokken_v1", "id", "'def'", "id = 'abc'"
        )
        _check_delete_permission_denied(
            engine, "level_b1", "gebieden_bouwblokken_v1", "id = 'abc'"
        )
        _check_truncate_permission_denied(engine, "level_b1", "gebieden_bouwblokken_v1")

    def test_auto_create_roles(self, here, engine, gebieden_schema_auth, dbsession, caplog):
        """
        Prove that dataset, table, and field permissions are set according,
        to the "OF-OF" Exclusief principle:

        * Een user met scope LEVEL/A mag alles uit de dataset gebieden zien,
          behalve tabel bouwblokken.
        * Een user met scope LEVEL/B mag alle velden van tabel bouwblokken zien,
          behalve beginGeldigheid.
        * Een user met scope LEVEL/C mag veld beginGeldigheid zien.

        Drie corresponderende users worden automatisch aangemaakt:
        'scope_level_a', 'scope_level_b', en 'scope_level_c;
        """

        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_auth, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {gebieden_schema_auth.id: gebieden_schema_auth}
        profile_path = here / "files" / "profiles" / "gebieden_test.json"
        with open(profile_path) as f:
            profile = json.load(f)
        profiles = {profile["name"]: profile}

        # These tests commented out due to: Error when trying to teardown test databases
        # Roles may still exist from previous test run. Uncomment when fixed:
        # _check_role_does_not_exist(engine, "scope_level_a")
        # _check_role_does_not_exist(engine, "scope_level_b")
        # _check_role_does_not_exist(engine, "scope_level_c")

        # Apply the permissions from Schema and Profiles.
        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine, "public", ams_schema, profiles, "AUTO", "ALL", create_roles=True, verbose=1
            )

        grants = _filter_grant_statements(caplog)
        assert grants == [
            "GRANT SELECT (begin_geldigheid) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_c",
            "GRANT SELECT (begingeldigheid) ON TABLE public.gebieden_ggwgebieden_v1 TO scope_level_a",
            "GRANT SELECT (eind_geldigheid) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_b",
            "GRANT SELECT (eindgeldigheid) ON TABLE public.gebieden_ggwgebieden_v1 TO scope_level_a",
            "GRANT SELECT (id) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_b",
            "GRANT SELECT (id) ON TABLE public.gebieden_ggwgebieden_v1 TO scope_level_a",
            "GRANT SELECT (identificatie) ON TABLE public.gebieden_ggwgebieden_v1 TO scope_level_a",
            "GRANT SELECT (ligt_in_buurt_id) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_d",
            "GRANT SELECT (ligt_in_buurt_identificatie) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_d",
            "GRANT SELECT (ligt_in_buurt_loose_id) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_d",
            "GRANT SELECT (ligt_in_buurt_volgnummer) ON TABLE public.gebieden_bouwblokken_v1 TO scope_level_d",
            "GRANT SELECT (volgnummer) ON TABLE public.gebieden_ggwgebieden_v1 TO scope_level_a",
            "GRANT SELECT ON SEQUENCE public.gebieden_bouwblokken_ligt_in_buurt_v1_id_seq TO scope_level_d",
            "GRANT SELECT ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO scope_level_a",
            "GRANT SELECT ON SEQUENCE public.gebieden_ggwgebieden_bestaat_uit_buurten_v1_id_seq TO scope_level_e",
            "GRANT SELECT ON SEQUENCE public.gebieden_ggwgebieden_gebieds_grenzen_v1_id_seq TO scope_level_f",
            "GRANT SELECT ON TABLE public.gebieden_bouwblokken_ligt_in_buurt_v1 TO scope_level_d",
            "GRANT SELECT ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO scope_level_a",
            "GRANT SELECT ON TABLE public.gebieden_buurten_v1 TO scope_level_a",
            "GRANT SELECT ON TABLE public.gebieden_ggwgebieden_bestaat_uit_buurten_v1 TO scope_level_e",
            "GRANT SELECT ON TABLE public.gebieden_ggwgebieden_gebieds_grenzen_v1 TO scope_level_f",
            "GRANT SELECT ON TABLE public.gebieden_wijken_v1 TO scope_level_a",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_ligt_in_buurt_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_bouwblokken_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_ligt_in_wijk_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_buurten_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_ggwgebieden_bestaat_uit_buurten_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_ggwgebieden_gebieds_grenzen_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_ggwgebieden_v1 TO write_gebieden",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.gebieden_wijken_v1 TO write_gebieden",
            "GRANT USAGE ON SEQUENCE public.gebieden_bouwblokken_ligt_in_buurt_v1_id_seq TO write_gebieden",
            "GRANT USAGE ON SEQUENCE public.gebieden_buurten_ligt_in_wijk_v1_id_seq TO write_gebieden",
            "GRANT USAGE ON SEQUENCE public.gebieden_ggwgebieden_bestaat_uit_buurten_v1_id_seq TO write_gebieden",
            "GRANT USAGE ON SEQUENCE public.gebieden_ggwgebieden_gebieds_grenzen_v1_id_seq TO write_gebieden",
        ]

        # Check if roles exist and the read priviliges are correct
        _check_select_permission_denied(engine, "scope_level_a", "gebieden_bouwblokken_v1")
        _check_select_permission_granted(engine, "scope_level_a", "gebieden_buurten_v1")

        _check_select_permission_granted(
            engine, "scope_level_b", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_denied(
            engine, "scope_level_b", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_denied(engine, "scope_level_b", "gebieden_buurten_v1")

        _check_select_permission_denied(
            engine, "scope_level_c", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "scope_level_c", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )
        _check_select_permission_granted(
            engine, "scope_level_d", "gebieden_bouwblokken_v1", "ligt_in_buurt_loose_id"
        )
        _check_select_permission_granted(
            engine, "scope_level_d", "gebieden_bouwblokken_v1", "ligt_in_buurt_id"
        )
        _check_select_permission_granted(
            engine, "scope_level_d", "gebieden_bouwblokken_v1", "ligt_in_buurt_identificatie"
        )
        _check_select_permission_granted(
            engine, "scope_level_d", "gebieden_bouwblokken_v1", "ligt_in_buurt_volgnummer"
        )
        _check_select_permission_denied(engine, "scope_level_c", "gebieden_buurten_v1")
        _check_select_permission_denied(engine, "scope_level_d", "gebieden_buurten_v1")

        # Check the through table, for all columns
        _check_select_permission_granted(
            engine, "scope_level_e", "gebieden_ggwgebieden_bestaat_uit_buurten_v1"
        )
        # Check the nested table, for all columns
        _check_select_permission_granted(
            engine, "scope_level_f", "gebieden_ggwgebieden_gebieds_grenzen_v1"
        )
        # Check the through table
        _check_select_permission_denied(
            engine, "scope_level_a", "gebieden_ggwgebieden_bestaat_uit_buurten_v1"
        )
        # Check the nested table
        _check_select_permission_denied(
            engine, "scope_level_a", "gebieden_ggwgebieden_gebieds_grenzen_v1"
        )

    def test_single_dataset_permissions(
        self, here, engine, gebieden_schema_auth, meetbouten_schema, dbsession
    ):
        """
        Prove when revoking grants on one dataset, other datasets are unaffected.
        """

        # dataset 1: gebieden
        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_auth, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("wijken", truncate=True, ind_extra_index=False)

        # dataset 2: meetbouten
        importer = NDJSONImporter(meetbouten_schema, engine)
        importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("metingen", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("referentiepunten", truncate=True, ind_extra_index=False)

        # Apply the permissions to gebieden
        apply_schema_and_profile_permissions(
            engine, "public", gebieden_schema_auth, None, "AUTO", "ALL", create_roles=True
        )
        # Check perms on gebieden
        _check_select_permission_granted(engine, "scope_level_a", "gebieden_buurten_v1")
        _check_select_permission_granted(
            engine, "scope_level_b", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )
        _check_select_permission_granted(
            engine, "scope_level_c", "gebieden_bouwblokken_v1", "begin_geldigheid"
        )

        # Apply the permissions to meetbouten
        apply_schema_and_profile_permissions(
            engine, "public", meetbouten_schema, None, "AUTO", "ALL", create_roles=True
        )
        # Check perms on meetbouten
        _check_select_permission_granted(engine, "scope_openbaar", "meetbouten_meetbouten_v1")

        # Revoke permissions for dataset gebieden and set grant again
        apply_schema_and_profile_permissions(
            engine,
            pg_schema="public",
            ams_schema=gebieden_schema_auth,
            profiles=None,
            role="AUTO",
            scope="ALL",
            create_roles=True,
            revoke=True,
        )
        # Check perms again on meetbouten
        _check_select_permission_granted(engine, "scope_openbaar", "meetbouten_meetbouten_v1")

    def test_permissions_support_shortnames(self, here, engine, hr_schema_auth, dbsession, caplog):
        """
        Prove that table, and field permissions are set on the shortnamed field.
        """

        ndjson_path = here / "files" / "data" / "hr_auth.ndjson"
        importer = NDJSONImporter(hr_schema_auth, engine)
        importer.generate_db_objects("sbiactiviteiten", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)

        # Setup schema and profile
        ams_schema = {hr_schema_auth.id: hr_schema_auth}

        # Apply the permissions from Schema and Profiles.
        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine,
                "public",
                ams_schema,
                None,
                "level_b",
                "LEVEL/B",
                create_roles=True,
                verbose=1,
            )
            apply_schema_and_profile_permissions(
                engine,
                "public",
                ams_schema,
                None,
                "level_c",
                "LEVEL/C",
                create_roles=True,
                verbose=1,
            )

        grants = _filter_grant_statements(caplog)
        assert grants == [
            "GRANT SELECT (identifier) ON TABLE public.hr_sbi_ac_v1 TO level_b",
            "GRANT SELECT (sbi_ac_naam) ON TABLE public.hr_sbi_ac_v1 TO level_b",
            "GRANT SELECT (sbi_ac_no) ON TABLE public.hr_sbi_ac_v1 TO level_c",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.hr_sbi_ac_v1 TO write_hr",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.hr_sbi_ac_v1 TO write_hr",
        ]

        # Check if the read priviliges are correct
        _check_select_permission_granted(engine, "level_b", "hr_sbi_ac_v1", "sbi_ac_naam")
        _check_select_permission_denied(engine, "level_b", "hr_sbi_ac_v1", "sbi_ac_no")
        _check_select_permission_denied(engine, "level_c", "hr_sbi_ac_v1", "sbi_ac_naam")
        _check_select_permission_granted(engine, "level_c", "hr_sbi_ac_v1", "sbi_ac_no")


class TestWritePermissions:
    def test_dataset_write_role(self, here, engine, gebieden_schema_auth):
        """
        Prove that a write role with name write_{dataset.id} is created with DML rights
        Check INSERT, UPDATE, DELETE, TRUNCATE permissions
        Check that for SELECT permissions you need an additional scope role.
        """

        ndjson_path = here / "files" / "data" / "gebieden.ndjson"
        importer = NDJSONImporter(gebieden_schema_auth, engine)
        importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)
        importer.generate_db_objects("buurten", truncate=True, ind_extra_index=False)

        # Setup schema
        ams_schema = {gebieden_schema_auth.id: gebieden_schema_auth}

        # The write_ roles do not have SELECT permissions
        _check_insert_permission_denied(
            engine, "write_gebieden", "gebieden_bouwblokken_v1", "id", "'abc'"
        )

        apply_schema_and_profile_permissions(
            engine=engine,
            pg_schema="public",
            ams_schema=ams_schema,
            profiles=None,
            role="AUTO",
            scope="ALL",
            set_read_permissions=True,
            set_write_permissions=True,
            create_roles=True,
            revoke=True,
        )

        # Drop testuser in case previous tests did not terminate correctly
        with engine.begin() as connection:
            connection.execute("DROP ROLE IF EXISTS testuser")

        _create_role(engine, "testuser")

        with engine.begin() as connection:
            connection.execute("GRANT write_gebieden TO testuser")

        #  It is now possible to INSERT data into the dataset tables
        _check_insert_permission_granted(
            engine, "testuser", "gebieden_bouwblokken_v1", "id", "'abc'"
        )

        #  The write_ roles do have SELECT permissions, therefore testuser should not have it
        _check_select_permission_granted(engine, "testuser", "gebieden_bouwblokken_v1")

        #  With SELECT it is possible to UPDATE or DELETE on given condition
        _check_update_permission_granted(
            engine, "testuser", "gebieden_bouwblokken_v1", "id", "'def'", "id = 'abc'"
        )
        _check_delete_permission_granted(
            engine, "testuser", "gebieden_bouwblokken_v1", "id = 'abc'"
        )

        # Add SELECT permissions by granting the appropriate scope to the user
        with engine.begin() as connection:
            connection.execute("GRANT scope_level_b TO testuser")

        # But now it's possible to SELECT the columns within scope level_b
        _check_select_permission_granted(
            engine, "testuser", "gebieden_bouwblokken_v1", "id, eind_geldigheid"
        )

        # And it's also possible to UPDATE and DELETE,
        # if the column for the condition is within scope
        _check_update_permission_granted(
            engine, "testuser", "gebieden_bouwblokken_v1", "id", "'def'", "id = 'abc'"
        )
        _check_delete_permission_granted(
            engine, "testuser", "gebieden_bouwblokken_v1", "id = 'def'"
        )

        # TRUNCATE is also allowed, even though the table is already empty by now
        _check_truncate_permission_granted(engine, "testuser", "gebieden_bouwblokken_v1")

    def test_multiple_datasets_write_roles(self, here, engine, parkeervakken_schema, afval_schema):
        """
        Prove that the write_{dataset.id} roles only have DML rights for their associated
        dataset tables.
        """

        importer = NDJSONImporter(parkeervakken_schema, engine)
        importer.generate_db_objects("parkeervakken", truncate=True, ind_extra_index=False)
        importer = NDJSONImporter(afval_schema, engine)
        importer.generate_db_objects("containers", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("clusters", truncate=True, ind_extra_index=False)

        # Setup schema and profile
        ams_schema = {afval_schema.id: afval_schema, parkeervakken_schema.id: parkeervakken_schema}

        apply_schema_and_profile_permissions(
            engine=engine,
            pg_schema="public",
            ams_schema=ams_schema,
            profiles=None,
            role="AUTO",
            scope="ALL",
            set_read_permissions=True,
            set_write_permissions=True,
            create_roles=True,
            revoke=True,
        )

        # Drop testuser in case previous tests did not terminate correctly
        with engine.begin() as connection:
            connection.execute("DROP ROLE IF EXISTS parkeer_tester")
            connection.execute("DROP ROLE IF EXISTS afval_tester")

        _create_role(engine, "parkeer_tester")
        _create_role(engine, "afval_tester")

        with engine.begin() as connection:
            connection.execute("GRANT write_parkeervakken TO parkeer_tester")
            connection.execute("GRANT write_afvalwegingen TO afval_tester")

        #  parkeer_tester has INSERT permission on parkeervakken datasets
        _check_insert_permission_granted(
            engine, "parkeer_tester", "parkeervakken_parkeervakken_v1", "id", "'abc'"
        )
        #  afval_tester has INSERT permission on afvalwegingen datasets
        _check_insert_permission_granted(
            engine, "afval_tester", "afvalwegingen_containers_v1", "id", "3"
        )
        #  parkeer_tester has NO INSERT permission on afvalwegingen datasets
        _check_insert_permission_denied(
            engine, "parkeer_tester", "afvalwegingen_containers_v1", "id", "3"
        )
        #  afval_tester has NO INSERT permission on parkeervakken datasets
        _check_insert_permission_denied(
            engine, "afval_tester", "parkeervakken_parkeervakken_v1", "id", "'abc'"
        )

    def test_permissions_support_shortnames(self, here, engine, hr_schema_auth, dbsession):
        """
        Prove that table, and field permissions are set on the shortnamed field.
        """

        ndjson_path = here / "files" / "data" / "hr_auth.ndjson"
        importer = NDJSONImporter(hr_schema_auth, engine)
        importer.generate_db_objects("sbiactiviteiten", truncate=True, ind_extra_index=False)
        importer.load_file(ndjson_path)

        # Setup schema and profile
        ams_schema = {hr_schema_auth.id: hr_schema_auth}

        # Apply the permissions from Schema and Profiles.
        apply_schema_and_profile_permissions(
            engine, "public", ams_schema, None, "AUTO", "ALL", create_roles=True
        )

        # Check if the write priviliges are correct
        _check_insert_permission_granted(
            engine,
            "write_hr",
            "hr_sbi_ac_v1",
            "sbi_ac_naam,sbi_ac_no,identifier",
            "'berry','14641','15101051'",
        )

    def test_setting_additional_grants(self, here, engine, meetbouten_schema, dbsession, caplog):
        """
        Prove that additional grants can be set using the extra argument.
        """

        importer = NDJSONImporter(meetbouten_schema, engine)
        importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("metingen", truncate=True, ind_extra_index=False)
        importer.generate_db_objects("referentiepunten", truncate=True, ind_extra_index=False)

        # Create the datasets_dataset table
        with engine.begin() as connection:
            connection.execute("CREATE TABLE datasets_dataset (id integer)")

        # Apply the permissions to meetbouten and add the extra grants to datasets_dataset
        with caplog.at_level(logging.INFO, logger="schematools.permissions.db"):
            apply_schema_and_profile_permissions(
                engine,
                "public",
                meetbouten_schema,
                None,
                "AUTO",
                "ALL",
                create_roles=True,
                verbose=1,
                additional_grants=("datasets_dataset:SELECT;scope_openbaar",),
            )

        grants = _filter_grant_statements(caplog)
        assert grants == [
            "GRANT SELECT ON SEQUENCE public.meetbouten_meetbouten_ligt_in_buurt_v1_id_seq TO scope_openbaar",
            "GRANT SELECT ON SEQUENCE public.meetbouten_metingen_refereertaanreferentiepunten_v1_id_seq TO scope_openbaar",
            "GRANT SELECT ON TABLE public.datasets_dataset TO scope_openbaar",
            "GRANT SELECT ON TABLE public.meetbouten_meetbouten_ligt_in_buurt_v1 TO scope_openbaar",
            "GRANT SELECT ON TABLE public.meetbouten_meetbouten_v1 TO scope_openbaar",
            "GRANT SELECT ON TABLE public.meetbouten_metingen_refereertaanreferentiepunten_v1 TO scope_openbaar",
            "GRANT SELECT ON TABLE public.meetbouten_metingen_v1 TO scope_openbaar",
            "GRANT SELECT ON TABLE public.meetbouten_referentiepunten_v1 TO scope_openbaar",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.meetbouten_meetbouten_ligt_in_buurt_v1 TO write_meetbouten",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.meetbouten_meetbouten_v1 TO write_meetbouten",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.meetbouten_metingen_refereertaanreferentiepunten_v1 TO write_meetbouten",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.meetbouten_metingen_v1 TO write_meetbouten",
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES ON TABLE public.meetbouten_referentiepunten_v1 TO write_meetbouten",
            "GRANT USAGE ON SEQUENCE public.meetbouten_meetbouten_ligt_in_buurt_v1_id_seq TO write_meetbouten",
            "GRANT USAGE ON SEQUENCE public.meetbouten_metingen_refereertaanreferentiepunten_v1_id_seq TO write_meetbouten",
        ]

        # Check perms on the datasets_dataset table
        _check_select_permission_granted(engine, "scope_openbaar", "datasets_dataset")


def _create_role(engine, role):
    """Create role. If role already exists just fail and ignore.
    This may happen if a previous pytest did not terminate correctly.
    """
    try:
        engine.execute(f'CREATE ROLE "{role}"')
    except ProgrammingError as e:
        if not isinstance(e.orig, DuplicateObject):
            raise


def _check_role_does_not_exist(engine, role):
    """Check if role does not exist"""
    with engine.begin() as connection:
        result = connection.execute("SELECT rolname FROM pg_roles WHERE rolname=%s", role)
        rows = list(result)
        assert len(rows) == 0


def _check_select_permission_denied(engine, role, table, column="*"):
    """Check if role has no SELECT permission on table.
    Fail if role, table or column does not exist.
    """
    with pytest.raises(
        Exception, match=f"permission denied for table {table}"
    ), engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        connection.execute(f"SELECT {column} FROM {table}")
        connection.execute("RESET ROLE")


def _check_select_permission_granted(engine, role, table, column="*"):
    """Check if role has SELECT permission on table.
    Fail if role, table or column does not exist.
    """
    with engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        result = connection.execute(f"SELECT {column} FROM {table}")
        connection.execute("RESET ROLE")
    assert result


def _check_insert_permission_granted(engine, role, table, column, value):
    """Check if role has INSERT permission on table.
    Fail if role, table or column does not exist, or value mismatches in datatype.
    """
    with engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        result = connection.execute(f"INSERT INTO {table} ({column}) VALUES ({value})")
        connection.execute("RESET ROLE")
    assert result


def _check_insert_permission_denied(engine, role, table, column, value):
    """Check if role has no INSERT permission on table.
    Fail if role, table or column does not exist.
    """
    with pytest.raises(Exception) as e_info, engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        connection.execute(f"INSERT INTO {table} ({column}) VALUES ({value})")
        connection.execute("RESET ROLE")
    assert f"permission denied for table {table}" in str(e_info)


def _check_update_permission_granted(engine, role, table, column, value, condition):
    """Check if role has UPDATE permission on table.
    Fail if role, table or column does not exist, or value mismatches in datatype.
    """
    with engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        result = connection.execute(f"UPDATE {table} SET {column} =  {value} WHERE {condition}")
        connection.execute("RESET ROLE")
    assert result


def _check_update_permission_denied(engine, role, table, column, value, condition):
    """Check if role has no UPDATE permission on table.
    Fail if role, table or column does not exist, or value mismatches in datatype.
    """
    with pytest.raises(Exception) as e_info, engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        connection.execute(f"UPDATE {table} SET {column} =  {value} WHERE {condition}")
        connection.execute("RESET ROLE")
    assert f"permission denied for table {table}" in str(e_info)


def _check_delete_permission_granted(engine, role, table, condition):
    """Check if role has DELETE permission on table.
    Fail if role, table or column does not exist, or value mismatches in datatype."""
    with engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        result = connection.execute(f"DELETE FROM {table} WHERE {condition}")  # noqa: S608
        connection.execute("RESET ROLE")
    assert result


def _check_delete_permission_denied(engine, role, table, condition):
    """Check if role has no DELETE permission on table.
    Fail if role, table or column does not exist, or value mismatches in datatype."""
    with pytest.raises(Exception) as e_info, engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        connection.execute(f"DELETE FROM {table} WHERE {condition}")  # noqa: S608
        connection.execute("RESET ROLE")
    assert f"permission denied for table {table}" in str(e_info)


def _check_truncate_permission_granted(engine, role, table):
    """Check if role has TRUNCATE permission on table.
    Fail if role or table does not exist.
    """
    with engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        result = connection.execute(f"TRUNCATE {table}")
        connection.execute("RESET ROLE")
    assert result


def _check_truncate_permission_denied(engine, role, table):
    """Check if role has no TRUNCATE permission on table.
    Fail if role or table does not exist.
    """
    with pytest.raises(Exception) as e_info, engine.begin() as connection:
        connection.execute(f"SET ROLE {role}")
        connection.execute(f"TRUNCATE {table}")
        connection.execute("RESET ROLE")
    assert f"permission denied for table {table}" in str(e_info)


def _filter_grant_statements(caplog):
    grants = sorted(
        m.replace("Executed --> ", "")
        for m in caplog.messages
        # Be specific in what is excluded, so unexpected notices can be detected.
        if not m.endswith('" already exists, skipping') and ("CREATE ROLE" not in m)
    )

    # Writes are seen multple times, because they use a single role.
    seen = set()
    seen_twice = {m for m in grants if (" TO write_" not in m) and (m in seen or seen.add(m))}
    newline = "\n"  # Python 3.10 f-string syntax doesn't support \
    assert not seen_twice, f"Duplicate grants: {newline.join(seen_twice)}"

    caplog.clear()
    return grants



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Iterator
from contextlib import closing
from functools import cached_property
from itertools import islice
from pathlib import Path
from typing import Any, Final, TypeVar

import click
import jsonpath_rw
import psycopg2
from psycopg2 import sql
from sqlalchemy import Boolean, exc, inspect, text
from sqlalchemy.dialects.postgresql.base import PGInspector
from sqlalchemy.engine.base import Engine
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.sql.ddl import CreateSchema
from sqlalchemy.sql.elements import TextClause
from sqlalchemy.sql.schema import Index, MetaData, Table

from schematools.factories import index_factory, tables_factory, views_factory
from schematools.types import DatasetSchema, DatasetTableSchema

metadata = MetaData()

#: We'd like to make a distinction between existing datasets that were originally, and
#: continue to be, created in the ``public`` PostgreSQL schema and newer datasets that will get
#: their own PostgreSQL schema. These newer datasets will also be versioned. Meaning that all
#: tables of these newer datasets will include their major and minor version in their name. Hence
#: the name of the SQL query to detect these newer datasets
IS_VERSIONED_DATASET_SQL: Final[TextClause] = text(
    """
    WITH public_dataset AS (  -- 'public' as in using public psql schema for a given dataset
        SELECT DISTINCT SPLIT_PART(TABLE_NAME, '_', 1) AS dataset_name
            FROM information_schema.tables
            WHERE table_schema = 'public'
              AND table_type = 'BASE TABLE'  -- specifically not 'VIEW'
            ORDER BY dataset_name),
         private_dataset AS (  -- 'private' as in using dataset specific psql schema
             SELECT schema_name AS dataset_name
                 FROM information_schema.schemata
                 WHERE schema_name NOT IN ('pg_toast',
                                           'pg_catalog',
                                           'public',
                                           'information_schema'
                     ))
    SELECT :dataset_name NOT IN (SELECT * FROM public_dataset) OR
           :dataset_name IN (SELECT * FROM private_dataset) AS is_versioned_dataset;
"""
).columns(is_versioned_dataset=Boolean)

T = TypeVar("T")


def chunked(stream: Iterator[T], size: int) -> Iterator[list[T]]:
    """Read parts of the generator, pause each time after a chunk."""

    # Based on more-itertools. islice returns results until 'size',
    # iter() repeatedly calls make_chunk until the '[]' sentinel is returned.
    def make_chunk():
        return list(islice(stream, size))

    return iter(make_chunk, [])


class JsonPathException(Exception):
    """Exception used to signal a jsonpath provenance."""


class Provenance:
    """Handler for the 'provenance' of a single field.
    This can resolve a field from an alias or JSONPath syntax.
    """

    def __init__(self, provenance: str):
        self.provenance: str = provenance
        self._json_path: jsonpath_rw.Child | None = (
            jsonpath_rw.parse(provenance) if provenance.startswith("$") else None
        )

    def __repr__(self):
        return f"Provenance({self.provenance!r})"

    def resolve(self, source: dict) -> Any:
        """Resolve provenance entries, return the value."""
        if self._json_path is not None:
            # JSONPath lookup
            matches = self._json_path.find(source)
            if not matches:
                raise LookupError(self.provenance)
            return matches[0].value
        else:
            # Alias name lookup
            return source[self.provenance]


class Record(dict):
    """A single generated row from the importer.
    Each dict key matches the database field name, each value is the data that should be written.

    This is a tiny wrapper around dict for simplicity that allows the source record
    to be accessed. The source record is useful debugging information,
    and some implementations need to retrieve non-schema fields
    (such as a "cursor" field in the last record).
    """

    def __init__(self, data, source: dict):
        super().__init__(data)
        self.source = source

    def __repr__(self):
        data_repr = super().__repr__()
        return f"Record({data_repr}, source={self.source!r})"


class BaseImporter:
    """Base importer that holds common data."""

    def __init__(
        self, dataset_schema: DatasetSchema, engine: Engine, logger: logging.Logger | None = None
    ) -> None:
        """Initializes the BaseImporter.

        dataset_schema: The dataset to work with.
        engine: SQLAlchemy database engine.
        """
        self.engine = engine
        self.dataset_schema = dataset_schema
        self.dataset_table: DatasetTableSchema | None = None
        self.db_table_name: str | None = None
        self.tables: dict[str, Table] = {}
        self.views: dict[str, sql.SQL] = {}
        self.pk_values_lookup: dict[str, set[Any]] = {}
        self.pk_colname_lookup: dict[str, str] = {}
        self.logger = LogfileLogger(logger) if logger else CliLogger()

    def deduplicate(
        self,
        table_name: str,
        table_records: list[Record],
    ) -> Iterator[Record]:
        """Removes duplicates from a set of records.
        This is introduced because the importer often needed to re-run
        before completing the task (e.g. source system errors / task redeploys).
        """
        pk_name = self.pk_colname_lookup.get(table_name)
        if pk_name is None:
            yield from table_records
            return

        # See which values already exist in the database. avoid reimporting them.
        values_lookup: set[Any] = self.pk_values_lookup.get(table_name, set())

        for record in table_records:
            value = record[pk_name]
            if value in values_lookup:
                self.logger.log_warning(
                    "Duplicate record for %s, with %s = %s", table_name, pk_name, value
                )
            else:
                yield record

            values_lookup.add(value)

    def create_pk_lookup(self, tables: dict[str, Table]) -> None:
        """Generate a lookup to avoid primary_key clashes."""
        for table_name, table in tables.items():
            pk_columns = inspect(table).primary_key.columns
            # nm tables do not have a PK
            if not pk_columns:
                continue

            # We assume a single PK (because of Django)
            pk_col = pk_columns.values()[0]
            pk_name = pk_col.name
            if pk_col.autoincrement:
                continue

            self.pk_colname_lookup[table_name] = pk_name
            with self.engine.connect() as conn:
                pks = {getattr(r, pk_name) for r in conn.execute(table.select())}

            self.pk_values_lookup[table_name] = pks

    def _schema_exists(self, schema_name: str) -> bool:
        """Check if a schema exists in the database."""
        with self.engine.connect() as conn:
            return bool(
                conn.scalar(
                    "SELECT EXISTS(SELECT 1 FROM pg_namespace WHERE nspname = %s)", schema_name
                )
            )

    def generate_db_objects(
        self,
        table_id: str,
        db_schema_name: str | None = None,
        db_table_name: str | None = None,
        truncate: bool = False,
        ind_tables: bool = True,
        ind_extra_index: bool = True,
        limit_tables_to: set | None = None,
        is_versioned_dataset: bool = False,
        ind_create_pk_lookup: bool = True,
    ) -> None:
        """Generate the tablemodels, tables and indexes.

        As default both table and index creation are set to True.

        Args:
            table_id: Name of the table as defined in the JSON schema defintion.
            db_schema_name: Name of the database schema where table should be
                created/present. Defaults to None (== public).
            db_table_name: Name of the table as defined in the database.
                Defaults to None.
            truncate: Indication to truncate table. Defaults to False.
            ind_tables: Indication to create table. Defaults to True.
            ind_extra_index: Indication to create indexes. Defaults to True.
            limit_tables_to: Only process the indicated tables. Normally, SA tables
                are generated for the whole dataset where `table_id` belongs to.
                Sometimes, this is not needed/wanted.
            is_versioned_dataset: Indicate whether the tables should be created in a private DB
                schema with a version in their name. See also:
                :attr:`.BaseImporter.is_versioned_dataset`. The private
                schema name will be derived from the dataset ID, unless overridden by the
                ``db_schema_name`` parameter.
        """
        self.dataset_table = self.dataset_schema.get_table_by_id(table_id)
        table_id = self.dataset_table.id  # get real-cased ID.

        # check if the dataset is a view
        if self.dataset_table.is_view:
            self.logger.log_info("Dataset %s is a view, skipping table generation", table_id)
            return False

        if is_versioned_dataset:
            if db_schema_name is None:
                # private DB schema instead of `public`
                db_schema_name = self.dataset_schema.id
            else:
                self.logger.log_warning(
                    "Versioning is specified, though schema name is explicitly overridden. "
                    "Is this really want you want?"
                )
            if db_table_name is None:
                db_table_name = self.dataset_table.db_name_variant(
                    # No dataset prefix as the tables will be created in their own
                    # private schema.
                    with_dataset_prefix=False,
                    with_version=True,
                )
            else:
                self.logger.log_warning(
                    "Versioning is specified, though table name is explicitly overridden. "
                    "Is this really what you want?"
                )

        if db_schema_name is not None and not self._schema_exists(db_schema_name):
            self.create_schema(db_schema_name)

        if ind_tables or ind_extra_index:
            if (dataset := self.dataset_table.dataset) is None:
                raise ValueError("Table {table_id} does not belong to a dataset")

            # Bind the metadata
            metadata.bind = self.engine

            # Get a table to import into
            self.tables = tables_factory(
                dataset,
                metadata=metadata,
                db_schema_names={table_id: db_schema_name},
                db_table_names={table_id: db_table_name},
                limit_tables_to=limit_tables_to,
                is_versioned_dataset=is_versioned_dataset,
            )

            if is_versioned_dataset:
                self.views = views_factory(dataset, self.tables)

        if db_table_name is None:
            db_table_name = self.dataset_table.db_name

        if ind_tables:
            self.prepare_tables(self.tables, truncate=truncate)
            if ind_create_pk_lookup:
                self.create_pk_lookup(self.tables)
            self.prepare_views()

        if ind_extra_index:
            # Get indexes to create
            indexes = index_factory(
                self.dataset_table,
                db_schema_name=db_schema_name,
                metadata=metadata,
                db_table_name=db_table_name,
                is_versioned_dataset=is_versioned_dataset,
            )
            metadata_inspector = inspect(metadata.bind)
            self.prepare_extra_index(
                indexes,
                metadata_inspector,
                metadata.bind,
                db_schema_name,
            )
            return None
        return None

    def table_exists(self, table_name: str) -> bool:
        """Check if a table exists in the database."""
        return table_name in self.engine.table_names()

    def create_view_user(self, table) -> None:
        """Create view users/owner in the database."""
        view_user = table.get_view_user()
        if view_user is not None:
            with closing(self.engine.raw_connection()) as conn, closing(conn.cursor()) as cur:
                cur.execute(f"CREATE USER {view_user}")
                conn.commit()

    def generate_view(self, dataset: DatasetSchema) -> None:
        """Generate a view for a schema."""
        raise NotImplementedError

    def load_file(self, file_name: Path, batch_size: int = 100, **kwargs: Any) -> Record | None:
        """Import a file into the database table, returns the last record, if available."""
        if self.dataset_table is None:
            raise ValueError("Import needs to be initialized with table info")
        data_generator = self.parse_records(
            file_name,
            self.dataset_table,
            **kwargs,
        )
        self.logger.log_start(file_name, size=batch_size)

        num_imported = 0
        insert_statements = {table_id: table.insert() for table_id, table in self.tables.items()}
        skipped_tables = set()

        last_record: Record | None = None
        for records in chunked(data_generator, size=batch_size):
            # every record is keyed on tablename + inside there is a list
            for table_id, table_records in self._group_records(records).items():
                try:
                    insert_statement = insert_statements[table_id]
                except KeyError:
                    if table_id not in skipped_tables:
                        # Show proper table db_name instead of confusing users with the internal ID
                        # If the resolving fails, the generator isn't producing proper table IDs
                        self.logger.log_info(
                            "Table '%s' was excluded, skipping!",
                            self.dataset_schema.get_table_by_id(table_id).db_name,
                        )
                        skipped_tables.add(table_id)
                    continue

                table_records = list(self.deduplicate(table_id, table_records))
                if table_records:
                    self.engine.execute(insert_statement, table_records)
            num_imported += len(records)
            self.logger.log_progress(num_imported)

            # Track the last record that's inserted for the main table.
            last_record = records[-1][self.dataset_table.id][0]

        self.logger.log_done(num_imported)
        return last_record

    def _group_records(self, records: list[dict[str, list[Record]]]) -> dict[str, list[Record]]:
        """Combine the records for a single table into a single set"""
        groups = defaultdict(list)
        for record in records:
            for table, rows in record.items():
                groups[table].extend(rows)

        return dict(groups)

    def parse_records(
        self, filename: Path, dataset_table: DatasetTableSchema, **kwargs: Any
    ) -> Iterator[dict[str, list[Record]]]:
        """Yield all records from the filename.
        The expected format of each returned row is::

            {
                "tableId1": [{"db_field1": "value", ...}, ...],
                "tableId2": [...],
            }
        """
        raise NotImplementedError()

    def prepare_tables(self, tables: dict[str, Table], truncate: bool = False) -> None:
        """Create the tables if needed."""
        for table in tables.values():
            if not table.exists():
                table.create()
            elif truncate:
                self.engine.execute(table.delete())

    def prepare_views(self) -> None:
        """Create views, if any."""
        if not self.views:
            return

        # sql.SQL requires an actual DBAPI connectionØ
        with closing(self.engine.raw_connection()) as conn, closing(conn.cursor()) as cur:
            for view in self.views.values():
                cur.execute(view)
            conn.commit()

    def prepare_extra_index(
        self,
        indexes: dict[str, list[Index]],
        inspector: PGInspector,
        engine: Engine,
        db_schema_name: str | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """Create extra indexes.

        Indexs are placed on identifiers columns in base tables
        and identifier columns in n:m tables, if not exists.
        """
        # setup logger
        _logger: CliLogger = LogfileLogger(logger) if logger else CliLogger()

        # create indexes, that do not exists yet
        for table_name, index_objects in indexes.items():
            try:
                db_indexes = inspector.get_indexes(table_name, schema=db_schema_name)
            except exc.NoSuchTableError:
                _logger.log_warning(f"Table '{table_name}' not found...skipping index creation")
                continue

            existing_db_indexes = {idx["name"] for idx in db_indexes}

            for index in index_objects:
                if index.name not in existing_db_indexes:
                    _logger.log_warning(f"Index '{index.name}' not found...creating")
                    index.create(bind=engine, checkfirst=True)

    @cached_property
    def is_versioned_dataset(self) -> bool:
        """Returns whether versioning will be employed for the current dataset.

        NB. Because the versioning has not been fully completed yet,
        this method always return False for now.


        Strictly speaking datasets are not directly (as in: on the dataset level) versioned
        anymore. Its tables, however, are! That is, in the Amsterdam Schema corresponding to the
        dataset. Whether we employ versioning on the DB level depends on whether we are dealing
        with:

            * an existing dataset that has been created in the ``public`` PostgreSQL schema
            * a dataset in a dataset specific PostgreSQL schema
            * a brand new dataset with no current DB representation.

        Versioning will be used for the latter two. Not for the first one."""
        return False

        # XXX Disabled until the decision has been made how to continue with versioning!

        # with self.engine.connect() as connection:
        #     is_versioned = cast(
        #         bool,
        #         connection.scalar(IS_VERSIONED_DATASET_SQL, dataset_name=self.dataset_schema.id),
        #     )
        # return is_versioned

    def create_schema(self, db_schema_name: str) -> None:
        """Create DB Schema.

        Is a no-op if schema already exists.
        """
        with self.engine.connect() as connection:
            try:
                connection.execute(CreateSchema(db_schema_name))
                self.logger.log_info("Created SQL schema %r", db_schema_name)
            except ProgrammingError as pe:
                if not isinstance(pe.orig, psycopg2.errors.DuplicateSchema):
                    # `CreateSchema` does not use the 'IF NOT EXISTS` clause.
                    # Hence we get an error if the schema already exists.
                    raise


class CliLogger:
    """Logger to be used when importer is called from the cli."""

    def log_start(self, file_name: Path, size: int) -> None:
        """Start the logging."""
        click.echo(
            f"Importing data from {file_name} [each dot is {size} records]: ", nl=False
        )  # noqa: T001

    def log_progress(self, num_imported: int) -> None:
        """Output progress."""
        click.echo(".", nl=False)  # noqa: T001

    def log_error(self, msg: str, *args: Any) -> None:
        """Output error."""
        click.echo(msg % args, err=True)  # noqa: T001

    def log_warning(self, msg: str, *args: Any) -> None:
        """Output warning."""
        click.echo(msg % args)  # noqa: T001

    def log_done(self, num_imported: int) -> None:
        """Indicate logging is finished."""
        click.echo(f" Done importing {num_imported} records")  # noqa: T001

    def log_info(self, msg: str, *args: Any) -> None:
        """Output informational message."""
        click.echo(msg % args)


class LogfileLogger(CliLogger):
    """Logger to be used when importer is called from python code."""

    def __init__(self, logger: logging.Logger):
        """Initialize logger."""
        self.logger = logger

    def log_start(self, file_name: Path, size: int) -> None:
        """Start the logging."""
        self.logger.info("Importing %s with %d records each:", file_name, size)

    def log_progress(self, num_imported: int) -> None:
        """Output progress."""
        self.logger.info("- imported %d records", num_imported)

    def log_error(self, msg: str, *args: Any) -> None:
        """Output error."""
        self.logger.error(msg, *args)

    def log_warning(self, msg: str, *args: Any) -> None:
        """Output warning."""
        self.logger.warning(msg, *args)

    def log_done(self, num_imported: int) -> None:
        """Indicate logging is finished."""
        self.logger.info("Done")

    def log_info(self, msg: str, *args: Any) -> None:
        """Output informational message."""
        self.logger.info(msg, *args)



from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from typing import IO

import psycopg2
from sqlalchemy import Column, MetaData, Table
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import ClauseElement

from schematools.factories import tables_factory
from schematools.types import _PUBLIC_SCOPE, DatasetFieldSchema, DatasetSchema, DatasetTableSchema

metadata = MetaData()


def enable_datetime_cast():
    """Register a special cast-to-string for the datetime type.

    This cast is not applied at module level, otherwise this
    cast would be enabled for all `schema-tools` code.
    """

    def cast_date(value, cur):
        if value is None:
            return None
        return str(value)

    # The bag_panden table had one record with a date (year = 0001)
    # that lets the dbapi `cursor.fetchmany` crash.
    # We need to define a custom type to handle this error,
    # because the historical record with this particular date cannot
    # be modified.
    date_oid = (1184,)  # This oid. can be obtained from cursor.description
    TIMESTAMPSTR = psycopg2.extensions.new_type(date_oid, "TIMESTAMP", cast_date)
    psycopg2.extensions.register_type(TIMESTAMPSTR)


class BaseExporter:
    """Baseclass for exporting tables rows."""

    extension = ""

    def __init__(
        self,
        connection: Connection,
        dataset_schema: DatasetSchema,
        output: str,
        table_ids: list[str] | None = None,
        scopes: list[str] | None = None,
        size: int | None = None,
        temporal_date: datetime | None = None,
    ):
        """Constructor.

        Args:
        connection: SQLAlchemy connection object.
        dataset_schema: Schema that needs export as geopackageself.
        output: path on the filesystem where output should be storedself.
        table_ids: optional parameter for a subset for the tables of the datasetself.
        scopes: Keycloak scopes that need to be taken into accountself.
            The geopackage will be produced contains information that is only
            accessible with these scopes.
        size: To produce a subset of the rows, mainly for testing.
        """
        self.connection = connection
        self.dataset_schema = dataset_schema
        self.table_ids = table_ids
        self.scopes = set(scopes)
        self.size = size
        self.temporal_date = temporal_date or datetime.now().astimezone()

        self.base_dir = Path(output)
        self.tables = (
            dataset_schema.tables
            if not table_ids
            else [dataset_schema.get_table_by_id(table_id) for table_id in table_ids]
        )
        self.sa_tables = tables_factory(dataset_schema, metadata)

    def _get_column(self, sa_table: Table, field: DatasetFieldSchema) -> Column:
        column = getattr(sa_table.c, field.db_name)
        # apply all processors
        for processor in self.processors:
            column = processor(field, column)

        return column

    def _get_columns(self, sa_table: Table, table: DatasetTableSchema) -> Iterable[Column]:
        for field in _get_fields(self.dataset_schema, table, self.scopes):
            try:
                yield self._get_column(sa_table, field)
            except AttributeError:
                pass  # skip unavailable columns

    def _get_temporal_clause(
        self, sa_table: Table, table: DatasetTableSchema
    ) -> ClauseElement | None:
        if not table.is_temporal:
            return None
        temporal = table.temporal
        for dimension in temporal.dimensions.values():
            start: Column = getattr(sa_table.c, dimension.start.db_name)
            end: Column = getattr(sa_table.c, dimension.end.db_name)
            return (
                # This is an SQLAlchemy statement, hence the &, | and == operators:
                (start <= self.temporal_date)
                & ((end > self.temporal_date) | (end == None))  # noqa: E711
            )
        return None

    def export_tables(self):
        for table in self.tables:
            srid = table.crs.split(":")[1] if table.crs else None
            if table.has_geometry_fields and srid is None:
                raise ValueError("Table has geo fields, but srid is None.")
            sa_table = self.sa_tables[table.id]
            columns = list(self._get_columns(sa_table, table))
            if not columns:
                continue
            with open(
                self.base_dir / f"{table.db_name}.{self.extension}", "w", encoding="utf8"
            ) as file_handle:
                self.write_rows(
                    file_handle,
                    table,
                    columns,
                    self._get_temporal_clause(sa_table, table),
                    srid,
                )

    def write_rows(  # noqa: D102
        self,
        file_handle: IO[str],
        table: DatasetTableSchema,
        columns: Iterable[Column],
        temporal_clause: ClauseElement | None,
        srid: str,
    ):
        raise NotImplementedError


def _get_fields(dataset_schema: DatasetSchema, table: DatasetTableSchema, scopes: list[str]):
    parent_scopes = set(dataset_schema.auth | table.auth) - {_PUBLIC_SCOPE}
    for field in table.fields:
        if field.is_array:
            continue
        if field.is_internal:
            continue
        if parent_scopes | set(field.auth) - {_PUBLIC_SCOPE} <= set(scopes):
            yield field



from __future__ import annotations

from pathlib import Path

from django.conf import settings
from django.core.management import BaseCommand
from django.db.models import Q

from schematools.contrib.django.models import Dataset
from schematools.loaders import FileSystemSchemaLoader, get_schema_loader
from schematools.naming import to_snake_case
from schematools.types import DatasetSchema

from .create_tables import create_tables
from .create_views import create_views


class Command(BaseCommand):
    help = """Import Amsterdam schema files.

    import_schemas imports all schemas from $SCHEMA_URL,
    unless schema files/URLs are given as positional arguments.
    """
    requires_system_checks = []

    def add_arguments(self, parser):
        parser.add_argument("schema", nargs="*", help="Local schema files to import")
        parser.add_argument(
            "--schema-url",
            default=settings.SCHEMA_URL,
            help=f"Schema URL (default: {settings.SCHEMA_URL})",
        )
        parser.add_argument("--create-tables", dest="create_tables", action="store_true")
        parser.add_argument("--create-views", dest="create_views", action="store_true")
        parser.add_argument("--no-create-tables", dest="create_tables", action="store_false")
        parser.set_defaults(create_tables=False)
        parser.set_defaults(create_views=False)

    def handle(self, *args, **options):
        if options["schema"]:
            datasets = self.import_from_files(options["schema"])
        else:
            datasets = self.import_from_url(options["schema_url"])

        if not datasets:
            self.stdout.write("No new datasets imported")
            return

        # Reasons for not creating tables directly are to manually configure the
        # "Datasets" model flags first. E.g. disable "enable_db", set a remote URL.
        if options["create_tables"]:
            create_tables(self, datasets, allow_unmanaged=True)

        if options["create_views"]:
            create_views(self, datasets)

    def import_from_files(self, schema_files) -> list[Dataset]:
        """Import all schema definitions from the given files."""
        datasets = []
        shared_loaders = {}

        def _get_shared_loader(path):
            # Make sure all datasets that share the same root also share the same loader,
            # so relations between the separate files can be resolved.
            if (loader := shared_loaders.get(path)) is None:
                shared_loaders[path] = loader = FileSystemSchemaLoader(
                    path, loaded_callback=self._loaded_callback
                )
            return loader

        for filename in schema_files:
            self.stdout.write(f"Loading schemas from {filename}")
            file = Path(filename).resolve()
            if file.is_dir():
                # As intended, read all datasets
                # The folder might be a sub path in the repository,
                # in which case only a few datasets will be imported.
                files = _get_shared_loader(file).get_all_datasets()
            else:
                # Previous logic also allowed selecting a single file by name.
                # This still needs to resolve the root to resolve relations,
                # and to make sure the 'path' is correctly calculated.
                root = FileSystemSchemaLoader.get_root(file)
                dataset = _get_shared_loader(root).get_dataset_from_file(file)
                # Random files may not follow the folder/dataset.json convention,
                # calculate the path here instead of using loader.get_dataset_path().
                path = file.parent.relative_to(root)
                if str(path) == ".":
                    path = dataset.id  # workaround for unit tests

                files = {path: dataset}

            datasets.extend(self._run_import(files))

        return datasets

    def import_from_url(self, schema_url) -> list[Dataset]:
        """Import all schema definitions from a URL"""
        self.stdout.write(f"Loading schema from {schema_url}")
        loader = get_schema_loader(schema_url, loaded_callback=self._loaded_callback)
        return self._run_import(loader.get_all_datasets())

    def _loaded_callback(self, schema: DatasetSchema):
        self.stdout.write(f"* Loaded {schema.id}")

    def _run_import(self, files: dict[str, DatasetSchema]) -> list[Dataset]:
        datasets = []
        for path, schema in files.items():
            self.stdout.write(f"* Processing {schema.id}")
            dataset = self._import(schema, path)
            if dataset is not None:
                datasets.append(dataset)

        return datasets

    def _import(self, schema: DatasetSchema, path: str) -> Dataset | None:
        """Import a single dataset schema."""
        try:
            dataset = Dataset.objects.get(name=Dataset.name_from_schema(schema))
        except Dataset.DoesNotExist:
            try:
                # try getting default dataset by name and version
                dataset = Dataset.objects.filter(Q(version=None) | Q(version=schema.version)).get(
                    name=to_snake_case(schema.id)
                )
            except Dataset.DoesNotExist:
                # Give up, Create new dataset
                dataset = Dataset.create_for_schema(schema, path)
                self.stdout.write(f"  Created {schema.id}")
                return dataset

        self.stdout.write(f"  Updated {schema.id}")
        if dataset.is_default_version != schema.is_default_version:
            self.update_dataset_version(dataset, schema)

        updated = dataset.save_for_schema(schema)
        dataset.save_path(path)
        return dataset if updated else None

    def update_dataset_version(self, dataset: Dataset, schema: DatasetSchema) -> None:
        """
        Perform dataset version update, including changes to dataset tables.
        """
        if not dataset.is_default_version and schema.is_default_version:
            try:
                current_default = Dataset.objects.get(name=Dataset.name_from_schema(schema))
            except Dataset.DoesNotExist:
                pass
            else:
                # Update current default dataset name to expected name.
                if current_default.version:
                    current_default.name = to_snake_case(f"{schema.id}_{current_default.version}")

                    current_default.save()

        dataset.name = Dataset.name_from_schema(schema)
        dataset.is_default_version = schema.is_default_version



from __future__ import annotations

from argparse import ArgumentParser
from typing import Any

from django.apps import AppConfig, apps
from django.core.management import BaseCommand

from schematools.contrib.django.models import DynamicModel


class Command(BaseCommand):
    """Dump the dynamically generated ModelFactories."""

    help = "Dump the (dynamic) ModelMocker definitions that Django holds in-memory."  # noqa: A003

    model_meta_args: list[tuple[str, Any]] = [
        # All possible options in Meta, with their defaults.
        # https://docs.djangoproject.com/en/3.2/ref/models/options/
        # The original model.Meta or Options.meta is not available after construction,
        # so will have to fill out the defaults to generate a reasonable representation.
        ("abstract", False),
        ("app_label", None),
        ("base_manager_name", None),
        ("constraints", []),
        ("db_table", None),
        ("default_manager_name", None),
        ("default_related_name", None),
        ("get_latest_by", None),
        ("index_together", ()),
        ("indexes", []),
        ("managed", True),
        ("order_with_respect_to", None),
        ("ordering", ["id"]),
        ("permissions", []),
        ("proxy", False),
        ("unique_together", ()),
        ("verbose_name", ""),  # is modelname
        ("verbose_name_plural", ""),  # is {modelname}s by default
    ]

    def add_arguments(self, parser: ArgumentParser) -> None:
        """Hook to add arguments."""
        parser.add_argument(
            "args",
            metavar="app_label",
            nargs="*",
            help="Names of Django apps to dump the ModelFactory for",
        )

    def handle(self, *args: str, **options: Any) -> None:
        """Main function of this command."""
        app_labels = sorted(args or apps.app_configs.keys())

        for app_label in app_labels:
            app: AppConfig = apps.get_app_config(app_label)
            self.write_header(app)
            models = app.get_models(include_auto_created=True)
            for model in sorted(models, key=lambda m: m._meta.model_name):
                # print(model)
                # <class 'dso_api.dynamic_api.aardgasvrijezones.models.buurt'>
                # <class 'dso_api.dynamic_api.aardgasvrijezones.models.buurtinitiatief'>
                self.write_model_mocker(model)

    def write_header(self, app: AppConfig) -> None:
        """Write app start header."""
        self.stdout.write(f"# ---- App: {app.verbose_name or app.label}\n\n\n")

    def write_model_mocker(self, model: type[DynamicModel]) -> None:
        """Write the representation of a complete DjangoModelFactory
        (ModelMocker in our terminology) to the output.
        """
        bases = ", ".join(base_class.__name__ for base_class in model.__bases__)
        self.stdout.write(f"class {model.__name__}({bases}):\n")

        for field in model._meta.get_fields():
            print(repr(field))
        # <django.db.models.fields.BigIntegerField: id>
        # <django.contrib.gis.db.models.fields.GeometryField: geometry>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurt_code>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurt_naam>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: toelichting>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: aandeel_kookgas>
        # class buurtinitiatief(DynamicModel):
        # <django.db.models.fields.BigIntegerField: id>
        # <django.contrib.gis.db.models.fields.GeometryField: geometry>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurtinitiatief_type>
        # <django_postgres_unlimited_varchar.UnlimitedCharField: buurt_naam>
        # <django.db.models.fields.FloatField: x_coordinaat>
        # <django.db.models.fields.FloatField: y_coordinaat>



from __future__ import annotations

import csv
from collections.abc import Iterable
from datetime import datetime
from typing import IO

from geoalchemy2 import functions as gfunc  # ST_AsEWKT
from sqlalchemy import Column, MetaData, func, select
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import ClauseElement

from schematools.exports import BaseExporter, enable_datetime_cast
from schematools.naming import toCamelCase
from schematools.types import DatasetFieldSchema, DatasetSchema, DatasetTableSchema

metadata = MetaData()


class CsvExporter(BaseExporter):  # noqa: D101
    extension = "csv"

    def geo_modifier(field: DatasetFieldSchema, column):
        if not field.is_geo:
            return column
        return gfunc.ST_AsEWKT(column).label(field.db_name)

    def id_modifier(field: DatasetFieldSchema, column):
        # We need an extra check for old relation definitions
        # that are defined as plain strings in the amsterdam schema
        # and not as objects.
        if field.table.is_temporal and (
            field.is_composite_key
            or field.related_table is not None
            and field.related_table.is_temporal
        ):
            return func.split_part(column, ".", 1).label(field.db_name)
        return column

    def datetime_modifier(field: DatasetFieldSchema, column):
        if field.type == "string" and field.format == "date-time":
            return func.to_char(column, 'YYYY-MM-DD"T"HH24:MI:SS').label(field.db_name)
        return column

    processors = (geo_modifier, id_modifier, datetime_modifier)

    def write_rows(  # noqa: D102
        self,
        file_handle: IO[str],
        table: DatasetTableSchema,
        columns: Iterable[Column],
        temporal_clause: ClauseElement | None,
        srid: str,
    ):

        field_names = [c.name for c in columns]
        writer = csv.DictWriter(file_handle, field_names, extrasaction="ignore")
        # Use capitalize() on headers, because csv export does the same
        writer.writerow({fn: toCamelCase(fn).capitalize() for fn in field_names})
        query = select(columns)
        if temporal_clause is not None:
            query = query.where(temporal_clause)
        if self.size is not None:
            query = query.limit(self.size)
        result = self.connection.execution_options(yield_per=1000).execute(query)
        for partition in result.partitions():
            for r in partition:
                writer.writerow(dict(r))


def export_csvs(
    connection: Connection,
    dataset_schema: DatasetSchema,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
    temporal_date: datetime | None = None,
):
    """Utility function to wrap the Exporter."""
    enable_datetime_cast()
    exporter = CsvExporter(
        connection, dataset_schema, output, table_ids, scopes, size, temporal_date
    )
    exporter.export_tables()
